"""
Compositely combine reader and writer classes for metadata/tags/properties operations on Track files.
"""
import datetime
from abc import ABCMeta, abstractmethod
from collections.abc import Generator
from copy import deepcopy, copy
from pathlib import Path
from typing import Any, Self

import mutagen
from yarl import URL

from musify.base import MusifyItem
from musify.exception import MusifyKeyError, MusifyAttributeError, MusifyTypeError, MusifyValueError
from musify.field import TagMap
from musify.field import TrackField
from musify.file.exception import FileDoesNotExistError, UnexpectedPathError
from musify.libraries.core.object import Track
from musify.libraries.local.base import LocalItem
# noinspection PyProtectedMember
from musify.libraries.local.track._tags import TagReader, TagWriter, SyncResultTrack
from musify.libraries.local.track.field import LocalTrackField as Tags
from musify.libraries.remote.core.wrangle import RemoteDataWrangler
from musify.types import UnitIterable
from musify.utils import to_collection


class LocalTrack[T: mutagen.FileType, U: TagReader, V: TagWriter](LocalItem, Track, metaclass=ABCMeta):
    """
    Generic track object for extracting, modifying, and saving metadata/tags/properties for a given file.

    :param file: The path or Mutagen object of the file to load.
    :param remote_wrangler: Optionally, provide a :py:class:`RemoteDataWrangler` object for processing URIs.
        This object will be used to check for and validate a URI tag on the file.
        The tag that is used for reading and writing is set by the ``uri_tag`` class attribute.
        If no ``remote_wrangler`` is given, no URI processing will occur.
    """

    __slots__ = (
        "_path",
        "_remote_wrangler",
        "_reader",
        "_writer",
        "_loaded",
        "_title",
        "_artist",
        "_album",
        "_album_artist",
        "_track_number",
        "_track_total",
        "_genres",
        "_year",
        "_month",
        "_day",
        "_bpm",
        "_key",
        "_disc_number",
        "_disc_total",
        "_compilation",
        "_comments",
        "_uri",
        "_has_uri",
        "_image_links",
        "_has_image",
        "_rating",
        "_date_added",
        "_last_played",
        "_play_count",
    )
    __attributes_classes__ = (Track, LocalItem)
    __attributes_ignore__ = ("tag_map",)

    @property
    def name(self):
        return self.title or self.filename

    @property
    @abstractmethod
    def tag_map(self) -> TagMap:
        """The map of tag names to tag IDs for the given file type."""
        raise NotImplementedError

    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, value: str | None):
        self._title = value

    @property
    def artist(self):
        return self._artist

    @artist.setter
    def artist(self, value: str | None):
        self._artist = value

    @property
    def artists(self) -> list[str]:
        return self._artist.split(self.tag_sep) if self._artist else []

    @artists.setter
    def artists(self, value: list[str]):
        self._artist = self.tag_sep.join(value)

    @property
    def album(self):
        return self._album

    @album.setter
    def album(self, value: str | None):
        self._album = value

    @property
    def album_artist(self):
        return self._album_artist

    @album_artist.setter
    def album_artist(self, value: str | None):
        self._album_artist = value

    @property
    def track_number(self):
        return self._track_number

    @track_number.setter
    def track_number(self, value: int | None):
        self._track_number = value

    @property
    def track_total(self):
        return self._track_total

    @track_total.setter
    def track_total(self, value: int | None):
        self._track_total = value

    @property
    def genres(self):
        return self._genres

    @genres.setter
    def genres(self, value: list[str] | None):
        self._genres = value

    @property
    def date(self):
        if self._year and self._month and self._day:
            return datetime.date(self._year, self._month, self._day)

    @date.setter
    def date(self, value: str | datetime.date | None):
        if isinstance(value, str):
            value = datetime.datetime.strptime(value, "%Y-%m-%d")
        self._year = value.year if value else None
        self._month = value.month if value else None
        self._day = value.day if value else None

    @property
    def year(self):
        return self._year

    @year.setter
    def year(self, value: int | None):
        if value and value < 1000:
            raise MusifyValueError(f"Year value is invalid: {value}")
        self._year = value

    @property
    def month(self):
        return self._month

    @month.setter
    def month(self, value: int | None):
        if value and (value < 1 or value > 12):
            raise MusifyValueError(f"Month value is invalid: {value}")
        self._month = value

    @property
    def day(self):
        return self._day

    @day.setter
    def day(self, value: int | None):
        if value and (value < 1 or value > 31):
            raise MusifyValueError(f"Day value is invalid: {value}")
        self._day = value

    @property
    def bpm(self):
        return self._bpm

    @bpm.setter
    def bpm(self, value: float | None):
        self._bpm = value

    @property
    def key(self):
        return self._key

    @key.setter
    def key(self, value: str | None):
        self._key = value

    @property
    def disc_number(self):
        return self._disc_number

    @disc_number.setter
    def disc_number(self, value: int | None):
        self._disc_number = value

    @property
    def disc_total(self):
        return self._disc_total

    @disc_total.setter
    def disc_total(self, value: int | None):
        self._disc_total = value

    @property
    def compilation(self):
        return self._compilation

    @compilation.setter
    def compilation(self, value: bool | None):
        self._compilation = value

    @property
    def comments(self):
        return self._comments

    @comments.setter
    def comments(self, value: UnitIterable[str] | None):
        self._comments = [value] if isinstance(value, str) else to_collection(value, list)

    def _uri_getter(self):
        return self._uri

    def _uri_setter(self, value: str | None):
        if value is None:
            self._uri = None
            self._has_uri = None
        elif self._reader.unavailable_uri_dummy is not None and value == self._reader.unavailable_uri_dummy:
            self._uri = None
            self._has_uri = False
        else:
            self._uri = value
            self._has_uri = True

        if self._loaded:
            setattr(self, self._reader.uri_tag.name.lower(), value)

    @property
    def has_uri(self):
        return self._has_uri

    @property
    def image_links(self):
        return self._image_links

    @image_links.setter
    def image_links(self, value: dict[str, str | Path | URL]):
        self._image_links = value

    @property
    def has_image(self):
        return self._has_image

    @has_image.setter
    def has_image(self, value: bool):
        self._has_image = value

    @property
    def length(self):
        return self._reader.file.info.length

    @property
    def rating(self):
        return self._rating

    @rating.setter
    def rating(self, value: float | None):
        self._rating = value

    @property
    def path(self):
        return self._path

    @property
    def type(self):
        """The type of audio file of this track"""
        return self.__class__.__name__

    @property
    def channels(self) -> int:
        """The number of channels in this audio file i.e. 1 for mono, 2 for stereo, ..."""
        return self._reader.file.info.channels

    @property
    def bit_rate(self) -> float:
        """The bit rate of this track in kilobytes per second"""
        return self._reader.file.info.bitrate / 1000

    @property
    def bit_depth(self) -> int | None:
        """The bit depth of this track in bits"""
        try:
            return self._reader.file.info.bits_per_sample
        except AttributeError:
            return None

    @property
    def sample_rate(self) -> float:
        """The sample rate of this track in kHz"""
        return self._reader.file.info.sample_rate / 1000

    @property
    def date_added(self) -> datetime.datetime | None:
        """The timestamp for when this track was added to the associated collection"""
        return self._date_added

    @date_added.setter
    def date_added(self, value: datetime.datetime | None):
        self._date_added = value

    @property
    def last_played(self) -> datetime.datetime | None:
        """The timestamp when this track was last played"""
        return self._last_played

    @last_played.setter
    def last_played(self, value: datetime.datetime | None):
        self._last_played = value

    @property
    def play_count(self) -> int | None:
        """The total number of times this track has been played"""
        return self._play_count

    @play_count.setter
    def play_count(self, value: int | None):
        self._play_count = value

    @abstractmethod
    def _create_reader(self, file: T) -> U:
        """Return a :py:class:`TagReader` object for this track type"""
        raise NotImplementedError

    @abstractmethod
    def _create_writer(self, file: T) -> V:
        """Return a :py:class:`TagWriter` object for this track type"""
        raise NotImplementedError

    def __init__(self, file: str | Path | T, remote_wrangler: RemoteDataWrangler = None):
        super().__init__()

        self._path: Path = Path(file if isinstance(file, str | Path) else file.filename)
        self._validate_type(self._path)

        self._remote_wrangler = remote_wrangler

        self._loaded = False
        self._reader: TagReader[T] | None = None
        self._writer: TagWriter[T] | None = None

        self._title = None
        self._artist = None
        self._album = None
        self._album_artist = None
        self._track_number = None
        self._track_total = None
        self._genres = None
        self._year = None
        self._month = None
        self._day = None
        self._bpm = None
        self._key = None
        self._disc_number = None
        self._disc_total = None
        self._compilation = None
        self._comments = None

        self._uri = None
        self._has_uri = None
        self._has_image = None
        self._image_links = {}

        self._rating = None
        self._date_added = None
        self._last_played = None
        self._play_count = None

        if isinstance(file, mutagen.FileType):
            self._reader = self._create_reader(file)
            self._writer = self._create_writer(file)
            self.refresh()

    def __await__(self) -> Generator[Any, None, Self]:
        return self.load().__await__()

    async def load(self) -> Self:
        """
        Load local file from scratch using mutagen from the path stored in this object.
        Refreshes the metadata loaded into this object and returns self.

        :return: Self.
        :raise FileDoesNotExistError: If the file cannot be found.
        :raise InvalidFileType: If the file type is not supported.
        """
        if not self._path.is_file():
            raise FileDoesNotExistError(self._path)

        file = mutagen.File(self.path)
        self._reader = self._create_reader(file)
        self._writer = self._create_writer(file)

        self.refresh()

        return self

    def refresh(self) -> None:
        """Extract update tags for this object from the loaded mutagen object."""
        self._refresh()

        # to reduce memory usage, remove any embedded images from the loaded file
        self._writer.clear_loaded_images()

        self._loaded = True

    def _refresh(self):
        """Simply refresh the metadata from file without any post-processing."""
        self.title = self._reader.read_title()
        self.artist = self._reader.read_artist()
        self.album = self._reader.read_album()
        self.album_artist = self._reader.read_album_artist()
        self.track_number = self._reader.read_track_number()
        self.track_total = self._reader.read_track_total()
        self.genres = self._reader.read_genres()
        self.year, self.month, self.day = self._reader.read_date() or (None, None, None)
        self.bpm = self._reader.read_bpm()
        self.key = self._reader.read_key()
        self.disc_number = self._reader.read_disc_number()
        self.disc_total = self._reader.read_disc_total()
        self.compilation = self._reader.read_compilation()
        self.comments = self._reader.read_comments()

        self.uri = self._reader.read_uri()
        if not self._loaded:
            self.has_image = self._reader.check_for_images()

    async def save(
            self, tags: UnitIterable[Tags] = Tags.ALL, replace: bool = False, dry_run: bool = True
    ) -> SyncResultTrack:
        """
        Update file's tags from given dictionary of tags.

        :param tags: Tags to be updated.
        :param replace: Destructively replace tags in each file.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: List of tags that have been updated.
        """
        # copy and reload the mutagen object to ensure comparison are being made against current data
        current = copy(self)
        current._writer.file.load(current._writer.file.filename)
        current._refresh()

        result = self._writer.write(source=current, target=self, tags=tags, replace=replace, dry_run=dry_run)

        # to reduce memory usage, remove any embedded images from the loaded file
        current._writer.clear_loaded_images()
        return result

    async def delete_tags(self, tags: UnitIterable[Tags] = (), dry_run: bool = True) -> SyncResultTrack:
        """
        Remove tags from file.

        :param tags: Tags to remove.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: List of tags that have been removed.
        """
        result = self._writer.delete_tags(tags=tags, dry_run=dry_run)
        if Tags.IMAGES in result.updated:
            self.has_image = False
        return result

    def merge(self, track: Track, tags: UnitIterable[TrackField] = TrackField.ALL) -> None:
        """Set the tags of this track equal to the given ``track``. Give a list of ``tags`` to limit which are set"""
        tag_names = TrackField.__tags__ if tags == TrackField.ALL else set(TrackField.to_tags(tags))

        for tag in tag_names:  # merge on each tag
            if hasattr(track, tag):
                setattr(self, tag, deepcopy(track[tag]))

    def extract_images_to_file(self, output_folder: str | Path) -> int:
        """Reload the file, extract and save all embedded images from file. Returns the number of images extracted."""
        self._reader.file.load(self._reader.file.filename)

        images = self._reader.read_images()
        if images is None:
            return 0

        output_folder = Path(output_folder)
        if not output_folder.is_dir():
            raise UnexpectedPathError(output_folder, "Given path must be a directory")

        count = 0
        for i, image in enumerate(images):
            output_path = output_folder.joinpath(self.filename + f"_{str(i).zfill(2)}")
            output_path = output_path.with_suffix("." + image.format.lower().lstrip("."))

            output_path.parent.mkdir(parents=True, exist_ok=True)
            image.save(output_path)
            count += 1

        # to reduce memory usage, remove any embedded images from the loaded file
        self._writer.clear_loaded_images()

        return count

    def as_dict(self):
        attributes_extra = {"remote_source": self._reader.remote_source}
        return self._get_attributes() | attributes_extra

    def __hash__(self):
        # If a class overrides __eq__, it must override __hash__ alongside it.
        # https://stackoverflow.com/questions/74664008/python-typeerror-unhashable-type-when-inheriting-from-subclass-with-hash
        # https://github.com/python/cpython/issues/46488
        return super().__hash__()

    def __eq__(self, item: MusifyItem):
        """Paths equal if both are LocalItems, URI attributes equal if both have a URI, names equal otherwise"""
        if hasattr(item, "path"):
            return self.path == item.path
        return super().__eq__(item)

    def __copy__(self):
        """Copy object by reloading from the file object in memory"""
        new = self.__class__(file=self.path, remote_wrangler=self._reader.remote_wrangler)
        new._reader = self._reader
        new._writer = self._writer

        if self._reader.file.tags:
            new._refresh()
            # image is deleted from loaded file on refresh in initial run
            # set parameter manually here rather than rely on the 2nd refresh to set it
            new.has_image = self.has_image
            new.image_links = self.image_links
        else:  # file is not a real file, used in testing. Set shallow copy of attributes manually
            keys = [key for key in LocalTrack.__slots__ if key.lstrip("_") in dir(self)]
            for key in keys:
                setattr(new, key, copy(getattr(self, key)))

        return new

    def __deepcopy__(self, _: dict = None):
        """Deepcopy object by reloading from the file on disk and refreshing loaded metadata from it."""
        new = self.__class__(file=self.path, remote_wrangler=self._reader.remote_wrangler)

        file = mutagen.File(self.path) if self._reader.file.tags else self._reader.file
        new._reader = self._create_reader(file)
        new._writer = self._create_writer(file)

        new.refresh()
        return new

    def __setitem__(self, key: str, value: Any):
        if not hasattr(self, key):
            raise MusifyKeyError(f"Given key is not a valid attribute of this item: {key}")

        attr = getattr(self, key)
        if isinstance(attr, property) and attr.fset is None:
            raise MusifyAttributeError(f"Cannot set values on the given key, it is protected: {key}")

        return setattr(self, key, value)

    def __or__(self, other: Track) -> Self:
        if not isinstance(other, Track):
            raise MusifyTypeError(
                f"Incorrect item given. Cannot merge with {other.__class__.__name__} as it is not a Track"
            )

        self_copy = copy(self)
        self_copy.merge(other, tags=TrackField.ALL)
        return self_copy

    def __ior__(self, other: Track) -> Self:
        if not isinstance(other, Track):
            raise MusifyTypeError(
                f"Incorrect item given. Cannot merge with {other.__class__.__name__} as it is not a Track"
            )

        self.merge(other, tags=TrackField.ALL)
        return self



"""
Implements all required non-items and non-playlist endpoints from the Spotify API.
"""
import logging
from abc import ABCMeta
from collections.abc import MutableMapping
from typing import Any

from musify.libraries.remote.core.types import RemoteIDType, RemoteObjectType
from musify.libraries.remote.spotify.api.base import SpotifyAPIBase
from musify.types import Number
from musify.utils import limit_value


class SpotifyAPIMisc(SpotifyAPIBase, metaclass=ABCMeta):

    __slots__ = ()

    async def print_collection(
            self,
            value: str | MutableMapping[str, Any] | None = None,
            kind: RemoteIDType | None = None,
            limit: int = 20,
    ) -> None:
        if not value:  # get user to paste in URL/URI
            value = input("\33[1mEnter URL/URI/ID: \33[0m")
        if not kind:
            kind = self.wrangler.get_item_type(value)
        key = self.collection_item_map[kind].name.lower()

        while kind is None:  # get user to input ID type
            kind = RemoteObjectType.from_name(input("\33[1mEnter ID type: \33[0m"))[0]

        id_ = self.wrangler.extract_ids(values=value, kind=kind)[0]
        url = self.wrangler.convert(id_, kind=kind, type_in=RemoteIDType.ID, type_out=RemoteIDType.URL)
        limit = limit_value(limit, floor=1, ceil=50)

        name = (await self.handler.get(url, params={"limit": limit}))["name"]
        response = await self.handler.get(f"{url}/{key}s", params={"limit": limit})

        i = 0
        while response.get("next") or i < response["total"]:  # loop through each page, printing data in blocks of 20
            if response.get("offset", 0) == 0:  # first page, show header
                url_ext = self.wrangler.convert(id_, kind=kind, type_in=RemoteIDType.ID, type_out=RemoteIDType.URL_EXT)
                self.logger.print_message(
                    f"\n\33[96mShowing tracks for {kind.name.lower()}\33[0m: "
                    f"\33[94m{name} \33[97m- {url_ext} \33[0m\n"
                )

            tracks = [item[key] if key in item else item for item in response[self.items_key]]
            for i, track in enumerate(tracks, i + 1):  # print each item in this page
                if isinstance(track["duration_ms"], Number):
                    length = track["duration_ms"] / 1000
                else:
                    length = track["duration_ms"]["totalMilliseconds"] / 1000
                self.print_item(i=i, name=track["name"], uri=track["uri"], length=length, total=response["total"])

            if response["next"]:
                response = await self.handler.get(response["next"], params={"limit": limit})
            self.logger.print_message()

    ###########################################################################
    ## GET endpoints
    ###########################################################################
    async def get_self(self, update_user_data: bool = True) -> dict[str, Any]:
        """
        ``GET: /me`` - Get API response for information on current user.

        :param update_user_data: When True, update the ``_user_data`` stored in this API object.
        """
        r = await self.handler.get(url=f"{self.url}/me")
        if update_user_data:
            self.user_data = r
        return r

    async def query(self, query: str | None, kind: RemoteObjectType, limit: int = 10) -> list[dict[str, Any]]:
        """
        ``GET: /search`` - Query for items. Modify result types returned with kind parameter

        :param query: Search query.
        :param kind: The remote object type to search for.
        :param limit: Number of results to get and return.
        :return: The response from the endpoint.
        """
        if not query or len(query) > 150:  # query is too short or too long, skip
            return []

        url = f"{self.url}/search"
        params = {'q': query, "type": kind.name.lower(), "limit": limit_value(limit, floor=1, ceil=50)}
        response = await self.handler.get(url, params=params)

        if "error" in response:
            self.handler.log("SKIP", url, message=[f"Query: {query}", response['error']], level=logging.ERROR)
            return []

        results = response[f"{kind.name.lower()}s"][self.items_key]
        if kind not in self.collection_item_map:
            return results

        key = self.collection_item_map[kind].name.lower() + "s"
        totals_key = {
            RemoteObjectType.ALBUM: "total_tracks",
            RemoteObjectType.SHOW: "total_episodes",
            RemoteObjectType.AUDIOBOOK: "total_chapters"
        }
        for result in results:
            if key in result and self.url_key in result[key] and "total" in result[key]:
                continue

            href = self.format_next_url(f"{result[self.url_key]}/{key}", limit=50)
            result[key] = {self.url_key: href, "total": result.get(totals_key.get(kind), 0)}

        return results



"""
Implements endpoints for getting items from the Spotify API.
"""
import re
from abc import ABCMeta
from collections.abc import Collection, Mapping, MutableMapping
from copy import copy
from itertools import batched
from typing import Any

from aiorequestful.types import URLInput
from yarl import URL

from musify.libraries.remote.core import RemoteResponse
from musify.libraries.remote.core.exception import APIError, RemoteObjectTypeError
from musify.libraries.remote.core.types import APIInputValueMulti, RemoteIDType, RemoteObjectType
from musify.libraries.remote.spotify.api.base import SpotifyAPIBase
from musify.utils import limit_value, to_collection

try:
    from tqdm.auto import tqdm
except ImportError:
    tqdm = None
ARTIST_ALBUM_TYPES = {"album", "single", "compilation", "appears_on"}


class SpotifyAPIItems(SpotifyAPIBase, metaclass=ABCMeta):

    __slots__ = ()

    _bar_threshold = 5

    def _get_unit(self, key: str | None = None, kind: str | None = None) -> str:
        """Determine the unit type to use in the progress bar"""
        if kind is None:
            kind = re.sub(r"[-_]+", " ", key) if key is not None else self.items_key
        return kind.lower().rstrip("s") + "s"

    ###########################################################################
    ## GET helpers: Generic methods for getting items
    ###########################################################################
    async def _get_items(
            self,
            url: URLInput,
            id_list: Collection[str],
            params: MutableMapping[str, Any] | None = None,
            key: str | None = None,
            kind: str | None = None,
            limit: int = 50,
    ) -> list[dict[str, Any]]:
        """
        Get responses from a given ``url`` appending an ID to each request from a given ``id_list``
        i.e. ``URL`` or ``ID``.

        When limit == 1, this function executes each URL request individually for each ID i.e. ``URL`` or ``ID``.
        Otherwise, it executes each URL request in batches of IDs based on the given ``limit`` size.
        It passes this chunked list of IDs to the request handler as a set of params in the form:
        ``{<ids>: '<comma separated string of IDs>'}``

        :param url: The base API URL endpoint for the required requests.
        :param id_list: List of IDs to append to the given URL.
        :param params: Extra parameters to add to each request.
        :param kind: The unit to use for logging. If None, determine from ``key`` or default to ``items``.
        :param key: The key to reference from each response to get the list of required values.
        :param limit: Size of each batch of IDs to get. This value will be limited to be between ``1`` and ``50``.
            When limit is 1, requests will be made individually without any batching params configured.
        :return: API JSON responses for each item at the given ``key``.
        :raise APIError: When the given ``key`` is not in the API response.
        """
        method = "GET"
        url = url.rstrip("/")
        kind = self._get_unit(key=key, kind=kind)
        params = params if params is not None else {}

        results, ids_cached, ids_not_cached = await self._get_responses_from_cache(
            method=method, url=url, id_list=id_list
        )

        if limit == 1:
            id_requests = ids_not_cached
        else:
            id_requests = list(batched(ids_not_cached, limit_value(limit, floor=1, ceil=50)))

        async def _get_result(i: int, id_: str | list[str]) -> dict[str, Any]:
            nonlocal params

            if isinstance(id_, str):  # single call
                href = f"{url}/{id_}"
                log = f"{kind.title()}: {len(ids_not_cached):>5}"
            else:  # batched call
                href = url
                params |= {"ids": ",".join(id_)}
                log = f"{kind.title() + ':':<11} {sum(map(len, id_requests[i:])):>6}/{len(ids_not_cached):<6}"

            response = await self.handler.request(
                method=method, url=href, params=params, persist=False, log_message=log
            )
            if key and key not in response:
                raise APIError(f"Given key {key!r} not found in response keys: {list(response.keys())}")

            if key:  # ensure identifiers are on each response so results can be sorted after execution
                for i, r in zip(id_, response[key], strict=True):
                    self._enrich_with_identifiers(response=r, id_=i, href=f"{url}/{i}")
            else:
                self._enrich_with_identifiers(response=response, id_=id_, href=f"{url}/{id_}")

            return response

        responses = await self.logger.get_asynchronous_iterator(
            (_get_result(i=i, id_=id_) for i, id_ in enumerate(id_requests)),
            desc=f"Getting {kind}",
            unit=kind if limit == 1 else "pages",
            disable=len(id_requests) < self._bar_threshold
        )
        if key:
            responses = [r for response in responses for r in response[key]]

        await self._cache_responses(method=method, responses=responses)

        id_list = tuple(id_list.keys()) if isinstance(id_list, Mapping) else to_collection(id_list)
        results.extend(responses)
        results.sort(key=lambda r: id_list.index(r[self.id_key]))

        return results

    async def extend_items(
            self,
            response: MutableMapping[str, Any] | RemoteResponse,
            kind: RemoteObjectType | str | None = None,
            key: RemoteObjectType | None = None,
            leave_bar: bool | None = None,
    ) -> list[dict[str, Any]]:
        """
        Extend the items for a given API ``response``.
        The function requests each page of the collection returning a list of all items
        found across all pages for this URL.

        Updates the value of the ``items`` key in-place by extending the value of the ``items`` key with new results.

        If a cache has been configured for this API, will also persist the items in any collection to the cache.

        If a :py:class:`RemoteResponse`, this function will not refresh it with the new response.
        The user must call `refresh` manually after execution.

        :param response: A remote API JSON response for an items type endpoint
            or a response/RemoteResponse which contains this response.
            Must include required keys:
            ``total`` and either ``next`` or ``href``, plus optional keys ``previous``, ``limit``, ``items`` etc.
        :param kind: The type of response being extended.
            If a RemoteObjectType is given, the method will attempt to enrich the items given
            and returned with given response on the key associated with this kind. The function will only do this
            if a parent response has been given and not an items block response.
        :param key: The type of response of the child objects. Used when selecting nested data for certain responses
            (e.g. user's followed artists).
        :param leave_bar: When a progress bar is displayed,
            toggle whether this bar should continue to be displayed after the operation is finished.
            When None, allow the logger to decide this setting.
        :return: API JSON responses for each item
        """
        if isinstance(response, RemoteResponse):
            response = response.response
        if not response:
            return []

        method = "GET"

        parent_key = kind
        parent_response = copy(response)

        key = self._format_key(key)
        response = response.get(key, response)
        if self.items_key not in response:
            response[self.items_key] = []

        if len(response[self.items_key]) == response["total"]:  # skip on fully extended response
            url = URL(response[self.url_key]).with_query(None)
            self.handler.log("SKIP", url, message="Response already extended")

            self._enrich_with_parent_response(
                response=response, key=key, parent_key=parent_key, parent_response=parent_response
            )
            return response[self.items_key]

        initial_url_key = "next" if response[self.items_key] and response.get("next") else self.url_key
        if not response.get(initial_url_key):
            return []

        initial_url = URL(response[initial_url_key])
        limit = int(initial_url.query.get("limit", 50))
        offset = int(initial_url.query.get("offset", len(response.get(self.items_key, []))))
        total = int(response["total"])

        urls = []
        for offset in range(offset, total, limit):
            params = dict(initial_url.query)
            params |= {"limit": limit, "offset": offset}
            urls.append(initial_url.with_query(params))

        async def _get_result(request: URL) -> dict[str, Any]:
            count = min(int(request.query["offset"]) + int(request.query["limit"]), response["total"])
            log = f"{count:>6}/{response["total"]:<6} {key or self.items_key}"
            r = await self.handler.request(method=method, url=request, log_message=log)
            return r.get(key, r)

        kind_name = self._format_key(kind) or self.items_key
        pages = (response["total"] - len(response[self.items_key])) / (response.get("limit", 1) or 1)
        results: list[dict[str, Any]] = await self.logger.get_asynchronous_iterator(
            map(_get_result, urls),
            initial=len(response[self.items_key]),
            total=response["total"],
            desc=f"Extending {kind_name}".rstrip("s") if kind_name[0].islower() else kind_name,
            unit=key or self.items_key,
            leave=leave_bar,
            disable=pages < self._bar_threshold,
        )
        results.sort(key=lambda r: r.get("offset", 0))  # tqdm doesn't execute in order, sort results

        # assign block values to response from last result
        final_result = next(reversed(results))
        response[self.url_key] = final_result[self.url_key]
        response["offset"] = final_result.get("offset")
        response["next"] = final_result.get("next")
        response["previous"] = final_result.get("previous")

        # assign results back to original response and enrich the child items
        response[self.items_key].extend(item for result in results for item in result[self.items_key])
        self._enrich_with_parent_response(
            response=response, key=key, parent_key=parent_key, parent_response=parent_response
        )

        # cache child items
        key = key.rstrip("s") if key else key
        results_to_cache = [
            result[key] if key and key in result else result for result in response[self.items_key]
        ]
        await self._cache_responses(method=method, responses=results_to_cache)

        return response[self.items_key]

    ###########################################################################
    ## Core GET endpoint methods
    ###########################################################################
    async def get_items(
            self,
            values: APIInputValueMulti[RemoteResponse],
            kind: RemoteObjectType | None = None,
            limit: int = 50,
            extend: bool = True,
    ) -> list[dict[str, Any]]:
        """
        ``GET: /{kind}s`` - Get information for given ``values``.

        ``values`` may be:
            * A string representing a URL/URI/ID.
            * A MutableSequence of strings representing URLs/URIs/IDs of the same type.
            * A remote API JSON response for a collection including:
                - some items under an ``items`` key,
                - a valid ID value under an ``id`` key,
                - a valid item type value under a ``type`` key if ``kind`` is None.
            * A MutableSequence of remote API JSON responses for a collection including the same structure as above.
            * A RemoteResponse of the appropriate type for this RemoteAPI which holds a valid API JSON response
              as described above.
            * A Sequence of RemoteResponses as above.

        If JSON response(s) given, this update each response given by merging with the new response
        and replacing the ``items`` with the new results.

        If :py:class:`RemoteResponse` values are given, this function will call `refresh` on them.

        :param values: The values representing some remote objects. See description for allowed value types.
            These items must all be of the same type of item i.e. all tracks OR all artists etc.
        :param kind: Item type if given string is ID.
        :param limit: When requests can be batched, size of batches to request.
            This value will be limited to be between ``1`` and ``50`` or ``20`` if getting albums.
        :param extend: When True and the given ``kind`` is a collection of items,
            extend the response to include all items in this collection.
        :return: API JSON responses for each item.
        :raise RemoteObjectTypeError: Raised when the function cannot determine the item type
            of the input ``values``. Or when it does not recognise the type of the input ``values`` parameter.
        """
        # input validation
        if not isinstance(values, RemoteResponse) and not values:  # skip on empty
            url = f"{self.url}/{self._format_key(kind)}" if kind else self.url
            self.handler.log("SKIP", url, message="No data given")
            return []
        if kind is None:  # determine the item type
            kind = self.wrangler.get_item_type(values)
        else:
            self.wrangler.validate_item_type(values, kind=kind)

        unit = self._format_key(kind)
        url = f"{self.url}/{unit}"
        id_list = self.wrangler.extract_ids(values, kind=kind)

        if kind in {RemoteObjectType.USER, RemoteObjectType.PLAYLIST} or len(id_list) <= 1:
            limit = 1  # force non-batched calls
        elif kind == RemoteObjectType.ALBUM:
            limit = limit_value(limit, floor=1, ceil=20)
        results = await self._get_items(
            url=url, id_list=id_list, kind=unit, key=unit if limit > 1 else None, limit=limit
        )

        key = self.collection_item_map.get(kind, kind)
        key_name = self._format_key(key)
        if len(results) == 0 or any(key_name not in result for result in results) or not extend:
            self._merge_results_to_input(original=values, responses=results, ordered=True)
            self._refresh_responses(responses=values, skip_checks=False)
            self.handler.log("DONE", url, message=f"Retrieved {len(results):>6} {unit}")
            return results

        async def _get_result(result: dict[str, Any]) -> None:
            self.handler.log("INFO", url, message=f"Extending {key_name} on {unit}")
            await self.extend_items(result, kind=kind, key=key, leave_bar=False)

        await self.logger.get_asynchronous_iterator(
            map(_get_result, results),
            desc=f"Extending {unit}",
            unit=unit,
            disable=len(id_list) < self._bar_threshold
        )

        self._merge_results_to_input(original=values, responses=results, ordered=True)
        self._refresh_responses(responses=values, skip_checks=False)

        item_count = sum(len(result[key_name][self.items_key]) for result in results)
        self.handler.log("DONE", url, message=f"Retrieved {item_count:>6} {key_name} across {len(results):>5} {unit}")

        return results

    async def get_user_items(
            self,
            user: str | None = None,
            kind: RemoteObjectType = RemoteObjectType.PLAYLIST,
            limit: int = 50,
    ) -> list[dict[str, Any]]:
        """
        ``GET: /{kind}s`` - Get saved items for a given user.

        :param user: The ID of the user to get playlists for. If None, use the currently authenticated user.
        :param kind: Item type to retrieve for the user.
            Spotify only supports ``PLAYLIST`` types for non-authenticated users.
        :param limit: Size of each batch of items to request in the collection items request.
            This value will be limited to be between ``1`` and ``50``.
        :return: API JSON responses for each collection.
        :raise RemoteIDTypeError: Raised when the input ``user`` does not represent a user URL/URI/ID.
        :raise RemoteObjectTypeError: Raised a user is given and the ``kind`` is not ``PLAYLIST``.
            Or when the given ``kind`` is not a valid collection.
        """
        # input validation
        if kind not in self.user_item_types:
            raise RemoteObjectTypeError(f"{kind.name.title()}s are not a valid user collection type", kind=kind)
        if kind != RemoteObjectType.PLAYLIST and user is not None:
            raise RemoteObjectTypeError(
                f"Only able to retrieve {kind.name.lower()}s from the currently authenticated user",
                kind=kind
            )

        params = {"limit": limit_value(limit, floor=1, ceil=50)}
        if user is not None:
            url = self.wrangler.convert(user, kind=RemoteObjectType.USER, type_out=RemoteIDType.URL)
            url = f"{url}/{kind.name.lower()}s"
            desc_qualifier = "user's"
        elif kind == RemoteObjectType.ARTIST:
            url = f"{self.url}/me/following"
            desc_qualifier = "current user's followed"
            params["type"] = "artist"
        else:
            url = f"{self.url}/me/{kind.name.lower()}s"
            desc_qualifier = "current user's" if kind == RemoteObjectType.PLAYLIST else "current user's saved"

        desc = f"Getting {desc_qualifier} {kind.name.lower()}s"
        initial = await self.handler.get(url, params=params)
        results = await self.extend_items(initial, kind=desc, key=kind)

        self.handler.log("DONE", url, message=f"Retrieved {len(results):>6} {kind.name.lower()}s")

        return results

    ###########################################################################
    ## Tracks GET endpoint methods
    ###########################################################################
    async def get_tracks(
            self,
            values: APIInputValueMulti[RemoteResponse],
            features: bool = False,
            analysis: bool = False,
            limit: int = 50,
            *_,
            **__,
    ) -> list[dict[str, Any]]:
        """
        ``GET: /{kind}s`` + ``GET: /audio-features`` and/or ``GET: /audio-analysis``

        Get track(s) info and any audio features/analysis.
        Mostly just a wrapper for ``get_items`` and ``extend_tracks`` functions.

        ``values`` may be:
            * A string representing a URL/URI/ID of type 'track'.
            * A MutableSequence of strings representing URLs/URIs/IDs of the type 'track'.
            * A remote API JSON response for a track including:
                - some items under an ``items`` key,
                - a valid ID value under an ``id`` key.
            * A MutableSequence of remote API JSON responses for a set of tracks including the same structure as above.
            * A RemoteResponse of the appropriate type for this RemoteAPI which holds a valid API JSON response
              as described above.
            * A Sequence of RemoteResponses as above.

        If JSON response(s) given, this updates each response given by adding the results
        under the ``audio_features`` and ``audio_analysis`` keys as appropriate.

        :param values: The values representing some remote track/s. See description for allowed value types.
        :param features: When True, get audio features.
        :param analysis: When True, get audio analysis.
        :param limit: Size of batches to request when getting audio features.
            This value will be limited to be between ``1`` and ``50``.
        :return: API JSON responses for each item, or the original response if the input ``values`` are API responses.
        :raise RemoteObjectTypeError: Raised when the item types of the input ``values`` are not all tracks or IDs.
        """
        tracks = await self.get_items(values=values, kind=RemoteObjectType.TRACK, limit=limit)

        # ensure that response are being assigned back to the original values if API response(s) given
        if isinstance(values, Mapping | RemoteResponse):
            tracks = [values]
        elif isinstance(values, Collection) and all(isinstance(v, Mapping | RemoteResponse) for v in values):
            tracks = values

        await self.extend_tracks(values=tracks, features=features, analysis=analysis, limit=limit)
        return tracks

    async def extend_tracks(
            self,
            values: APIInputValueMulti[RemoteResponse],
            features: bool = False,
            analysis: bool = False,
            limit: int = 50,
    ) -> list[dict[str, Any]]:
        """
        ``GET: /audio-features`` and/or ``GET: /audio-analysis`` - Get audio features/analysis for given track/s.

        ``values`` may be:
            * A string representing a URL/URI/ID of type 'track'.
            * A MutableSequence of strings representing URLs/URIs/IDs of the type 'track'.
            * A remote API JSON response for a track including:
                - some items under an ``items`` key,
                - a valid ID value under an ``id`` key.
            * A MutableSequence of remote API JSON responses for a set of tracks including the same structure as above.
            * A RemoteResponse of the appropriate type for this RemoteAPI which holds a valid API JSON response
              as described above.
            * A Sequence of RemoteResponses as above.

        If JSON response(s) given, this updates each response given by adding the results
        under the ``audio_features`` and ``audio_analysis`` keys as appropriate.

        If :py:class:`RemoteResponse` values are given, this function will call `refresh` on them.

        :param values: The values representing some remote track/s. See description for allowed value types.
        :param features: When True, get audio features.
        :param analysis: When True, get audio analysis.
        :param limit: Size of batches to request when getting audio features.
            This value will be limited to be between ``1`` and ``50``.
        :return: API JSON responses for each item.
            Mapped to ``audio_features`` and ``audio_analysis`` keys as appropriate.
        :raise RemoteObjectTypeError: Raised when the item types of the input ``values`` are not all tracks or IDs.
        """
        # input validation
        if not features and not analysis:  # skip on all False
            return []
        if not values:  # skip on empty
            self.handler.log("SKIP", self.url, message="No data given")
            return []
        self.wrangler.validate_item_type(values, kind=RemoteObjectType.TRACK)

        id_list = self.wrangler.extract_ids(values, kind=RemoteObjectType.TRACK)

        # value list takes the form [URL, key, batched]
        config: dict[str, tuple[str, str, bool]] = {}
        if features:
            config["features"] = (f"{self.url}/audio-features", "audio_features", True)
        if analysis:
            config["analysis"] = (f"{self.url}/audio-analysis", "audio_analysis", False)

        async def _get_result(kind: str, url: str, key: str, batch: bool) -> dict[str, list[dict[str, Any]]]:
            _key = key
            _limit = limit
            if not batch or len(id_list) <= 1:
                _key = None
                _limit = 1

            return {key: await self._get_items(url=url, id_list=id_list, kind=kind, key=_key, limit=_limit)}

        results: list[dict[str, Any]] = []
        bar = self.logger.get_asynchronous_iterator(
            (_get_result(kind=kind, url=url, key=key, batch=batch) for kind, (url, key, batch) in config.items()),
            disable=True,
        )
        for result_map in await bar:
            for key, responses in result_map.items():
                responses.sort(key=lambda response: id_list.index(response[self.id_key]))
                responses = ({self.id_key: response[self.id_key], key: response} for response in responses)
                results = list(responses) if not results \
                    else [rs | rp for rs, rp in zip(results, responses, strict=True)]

        self._merge_results_to_input(original=values, responses=results, ordered=False, clear=False)
        self._refresh_responses(responses=values, skip_checks=False)

        def map_key(value: str) -> str:
            """Map the given ``value`` to logging appropriate string"""
            return value.replace("_", " ").replace("analysis", "analyses")

        log_types = " and ".join(map_key(key) for _, key, _ in config.values())
        self.handler.log(
            method="DONE",
            url=f"{self.url}/{"+".join(c[0].split("/")[-1] for c in config.values())}",
            message=f"Retrieved {log_types} for {len(id_list):>5} tracks"
        )

        return results

    ###########################################################################
    ## Artists GET endpoints methods
    ###########################################################################
    async def get_artist_albums(
            self, values: APIInputValueMulti[RemoteResponse], types: Collection[str] = (), limit: int = 50,
    ) -> dict[str, list[dict[str, Any]]]:
        """
        ``GET: /artists/{ID}/albums`` - Get all albums associated with the given artist/s.

        ``values`` may be:
            * A string representing a URL/URI/ID of type 'artist'.
            * A MutableSequence of strings representing URLs/URIs/IDs of the type 'artist'.
            * A remote API JSON response for an artist including a valid ID value under an ``id`` key.
            * A MutableSequence of remote API JSON responses for a set of artists including the same structure as above.
            * A RemoteResponse of the appropriate type for this RemoteAPI which holds a valid API JSON response
              as described above.
            * A Sequence of RemoteResponses as above.

        If JSON response(s) given, this updates each response given by adding the results under the ``albums`` key.

        If :py:class:`RemoteResponse` values are given, this function will call `refresh` on them.

        :param values: The values representing some remote artist/s. See description for allowed value types.
        :param types: The types of albums to return. Select from ``{"album", "single", "compilation", "appears_on"}``.
        :param limit: Size of batches to request.
            This value will be limited to be between ``1`` and ``50``.
        :return: A map of the Artist ID to a list of the API JSON responses for each album.
        :raise RemoteObjectTypeError: Raised when the item types of the input ``values`` are not all artists or IDs.
        """
        url = f"{self.url}/artists/{{id}}/albums"

        # input validation
        if not isinstance(values, RemoteResponse) and not values:  # skip on empty
            self.handler.log("SKIP", url, message="No data given")
            return {}

        if types and not all(t in ARTIST_ALBUM_TYPES for t in types):
            raise APIError(
                f"Given types not recognised, must be one or many of the following: {ARTIST_ALBUM_TYPES} ({types})"
            )
        self.wrangler.validate_item_type(values, kind=RemoteObjectType.ARTIST)

        params = {"limit": limit_value(limit, floor=1, ceil=50)}
        if types:
            params["include_groups"] = ",".join(set(types))

        key = RemoteObjectType.ALBUM
        results: dict[str, dict[str, Any]] = {}

        async def _get_result(id_: str) -> None:
            results[id_] = await self.handler.get(url=url.format(id=id_), params=params)
            await self.extend_items(results[id_], kind="artist albums", key=key, leave_bar=False)

        id_list = self.wrangler.extract_ids(values, kind=RemoteObjectType.ARTIST)
        await self.logger.get_asynchronous_iterator(
            map(_get_result, id_list),
            desc="Getting artist albums",
            unit="artist",
            disable=len(id_list) < self._bar_threshold
        )

        for result in results.values():  # add skeleton items block to album responses
            for album in result[self.items_key]:
                album["tracks"] = {
                    self.url_key: self.format_next_url(
                        url=str(URL(album[self.url_key]).with_query(None)) + "/tracks", offset=0, limit=50
                    ),
                    "total": album["total_tracks"]
                }

        results_remapped = [{self.id_key: id_, "albums": result} for id_, result in results.items()]
        self._merge_results_to_input(original=values, responses=results_remapped, ordered=False, clear=False)
        self._refresh_responses(responses=values, skip_checks=True)

        item_count = sum(map(len, results.values()))
        self.handler.log(
            method="DONE",
            url=url.format(id="..."),
            message=f"Retrieved {item_count:>6} albums across {len(results):>5} artists",
        )

        return {k: v[self.items_key] for k, v in results.items()}



"""
Core exceptions for the entire package.
"""
from typing import Any


class MusifyError(Exception):
    """Generic base class for all Musify-related errors"""


class MusifyKeyError(MusifyError, KeyError):
    """Exception raised for invalid keys."""


class MusifyValueError(MusifyError, ValueError):
    """Exception raised for invalid values."""


class MusifyTypeError(MusifyError, TypeError):
    """Exception raised for invalid types."""
    def __init__(self, kind: Any, message: str = "Invalid item type given"):
        self.message = message
        super().__init__(f"{self.message}: {kind}")


class MusifyAttributeError(MusifyError, AttributeError):
    """Exception raised for invalid attributes."""


class MusifyImportError(MusifyError, ImportError):
    """Exception raised for import errors, usually from missing modules."""


###########################################################################
## Enum errors
###########################################################################
class MusifyEnumError(MusifyError):
    """
    Exception raised for errors related to :py:class:`MusifyEnum` implementations.

    :param value: The value that caused the error.
    :param message: Explanation of the error.
    """

    def __init__(self, value: Any, message: str = "Could not find enum"):
        self.message = message
        super().__init__(f"{self.message}: {value}")


class FieldError(MusifyEnumError):
    """
    Exception raised for errors related to :py:class:`Field` enums.

    :param message: Explanation of the error.
    """
    def __init__(self, message: str | None = None, field: Any | None = None):
        super().__init__(value=field, message=message)



"""
The M3U implementation of a :py:class:`LocalPlaylist`.
"""
import asyncio
from collections.abc import Collection
from dataclasses import dataclass
from pathlib import Path
from typing import Self

from musify.base import Result
from musify.file.base import File
from musify.file.path_mapper import PathMapper
from musify.libraries.local.playlist.base import LocalPlaylist
from musify.libraries.local.track import LocalTrack, load_track
from musify.libraries.remote.core.wrangle import RemoteDataWrangler
from musify.processors.filter import FilterDefinedList


@dataclass(frozen=True)
class SyncResultM3U(Result):
    """Stores the results of a sync with a local M3U playlist"""
    #: The total number of tracks in the playlist before the sync.
    start: int
    #: The number of tracks added to the playlist.
    added: int
    #: The number of tracks removed from the playlist.
    removed: int
    #: The number of tracks that were in the playlist before and after the sync.
    unchanged: int
    #: The difference between the total number tracks in the playlist from before and after the sync.
    difference: int
    #: The total number of tracks in the playlist after the sync.
    final: int


class M3U(LocalPlaylist[FilterDefinedList[str | Path | File]]):
    """
    For reading and writing data from M3U playlist format.

    :param path: Absolute path of the playlist.
        If the playlist ``path`` given does not exist, a new playlist will be created on :py:meth:`save`
    :param path_mapper: Optionally, provide a :py:class:`PathMapper` for paths stored in the playlist file.
        Useful if the playlist file contains relative paths and/or paths for other systems that need to be
        mapped to absolute, system-specific paths to be loaded and back again when saved.
    :param remote_wrangler: Optionally, provide a :py:class:`RemoteDataWrangler` object for processing URIs on tracks.
        If given, the wrangler can be used when calling __get_item__ to get an item from the collection from its URI.
        The wrangler is also used when loading tracks to allow them to process URI tags.
        For more info on this, see :py:class:`LocalTrack`.
    """

    __slots__ = ("_description",)

    valid_extensions = frozenset({".m3u"})

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value: str | None):
        self._description = value

    @property
    def image_links(self):
        return {}

    def __init__(
            self,
            path: str | Path,
            path_mapper: PathMapper = PathMapper(),
            remote_wrangler: RemoteDataWrangler = None,
    ):
        super().__init__(path=path, path_mapper=path_mapper, remote_wrangler=remote_wrangler)

        self._description = None

    async def _load_track(self, path: str | Path) -> LocalTrack:
        path = self.path_mapper.map(path, check_existence=True)
        return await load_track(path=path, remote_wrangler=self.remote_wrangler)

    async def load(self, tracks: Collection[LocalTrack] = ()) -> Self:
        """
        Read the playlist file and update the tracks in this playlist instance.

        :param tracks: Available Tracks to search through for matches.
            If no tracks are given, the playlist instance will load all the tracks
            from scratch according to its settings.
        :return: Self
        """
        path_list: list[Path] = []
        if self.path.is_file():  # load from file
            with open(self.path, "r", encoding="utf-8") as file:
                paths_raw = self.path_mapper.map_many([line.strip() for line in file], check_existence=True)
            path_list = list(map(Path, paths_raw))

            if not path_list:  # empty playlist file
                self.clear()
                self._original = []
                return self

        self.matcher = FilterDefinedList(values=path_list)
        self.matcher.transform = lambda x: Path(self.path_mapper.map(x, check_existence=False))

        if tracks:  # match paths from given tracks using the matcher
            self._match(tracks)
        else:  # use the paths in the matcher to load tracks from scratch
            self.tracks = await asyncio.gather(
                *map(self._load_track, filter(lambda path: path is not None, self.matcher.values))
            )

        self._limit(ignore=self.matcher.values)
        self._sort()

        self._original = self.tracks.copy() if self._path.is_file() else []

        return self

    async def save(self, dry_run: bool = True, *_, **__) -> SyncResultM3U:
        """
        Write the tracks in this Playlist and its settings (if applicable) to file.

        :param dry_run: Run function, but do not modify the file on the disk.
        :return: The results of the sync as a :py:class:`SyncResultM3U` object.
        """
        start_paths = set(map(Path, self.path_mapper.unmap_many(self._original, check_existence=False)))
        self.path.parent.mkdir(parents=True, exist_ok=True)

        if not dry_run:
            with open(self.path, "w", encoding="utf-8") as file:
                # reassign any original folder found by the matcher and output
                paths = self.path_mapper.unmap_many(self.tracks, check_existence=False)
                file.writelines(path.strip() + '\n' for path in paths)

            self._original = self.tracks.copy()  # update original tracks to newly saved tracks

            with open(self.path, "r", encoding="utf-8") as file:  # get list of paths that were saved for results
                final_paths = {Path(line.rstrip()) for line in file if line.rstrip()}
        else:  # use current list of tracks as a proxy of paths that were saved for results
            final_paths = set(map(Path, self.path_mapper.unmap_many(self._tracks, check_existence=False)))

        return SyncResultM3U(
            start=len(start_paths),
            added=len(final_paths - start_paths),
            removed=len(start_paths.difference(final_paths)),
            unchanged=len(start_paths.intersection(final_paths)),
            difference=len(final_paths) - len(start_paths),
            final=len(final_paths),
        )



from http import HTTPMethod
from typing import Any

from aiorequestful.cache.backend.base import ResponseRepositorySettings
from aiorequestful.types import MethodInput, URLInput
from yarl import URL

from musify.libraries.remote.core.exception import RemoteObjectTypeError
from musify.libraries.remote.core.types import RemoteIDType
from musify.libraries.remote.spotify.wrangle import SpotifyDataWrangler


class SpotifyRepositorySettings(ResponseRepositorySettings):

    @property
    def fields(self) -> tuple[str, ...]:
        return "id",

    def get_key(self, method: MethodInput, url: URLInput, **__) -> tuple[str | None, ...]:
        if HTTPMethod(method) != HTTPMethod.GET:
            return (None,)

        try:
            return SpotifyDataWrangler.convert(str(url), type_in=RemoteIDType.URL, type_out=RemoteIDType.ID),
        except RemoteObjectTypeError:
            pass
        return (None,)

    def get_name(self, payload: dict[str, Any]) -> str | None:
        if payload.get("type") == "user":
            return payload["display_name"]
        return payload.get("name")


class SpotifyPaginatedRepositorySettings(SpotifyRepositorySettings):

    @property
    def fields(self) -> tuple[str, ...]:
        return *super().fields, "offset", "size"

    def get_key(self, method: MethodInput, url: URLInput, **__) -> tuple[str | int | None, ...]:
        base = super().get_key(method=method, url=url)
        return *base, self.get_offset(url), self.get_limit(url)

    @staticmethod
    def get_offset(url: URLInput) -> int:
        """Extracts the offset for a paginated request from the given ``url``."""
        params = URL(url).query
        return int(params.get("offset", 0))

    @staticmethod
    def get_limit(url: URLInput) -> int:
        """Extracts the limit for a paginated request from the given ``url``."""
        params = URL(url).query
        return int(params.get("limit", 50))



"""
All core :py:class:`Field` implementations relating to
core :py:class:`MusifyItem` and :py:class`MusifyCollection` implementations.
"""
from collections.abc import Sequence
from dataclasses import field, dataclass
from typing import Self

from musify.types import UnitIterable, MusifyEnum


class Field(MusifyEnum):
    """Base class for field names of an item."""

    @classmethod
    def from_name(cls, *names: str) -> list[Self]:
        """
        Returns all enums that match the given enum names

        :raise EnumNotFoundError: If a corresponding enum cannot be found.
        """
        return super().from_name(*names, fail_on_many=False)

    @classmethod
    def from_value(cls, *values: int) -> list[Self]:
        """
        Returns all enums that match the given enum values

        :raise EnumNotFoundError: If a corresponding enum cannot be found.
        """
        return super().from_value(*values, fail_on_many=False)


class Fields(Field):
    """
    All possible Field enums in this program.

    This is used to ensure all Field enum implementations have the same values for their enum names.
    """
    # all numbers relate to the number of the field value as mapped by MusicBee
    # changing these numbers will break all MusicBee functionality unless it is noted
    # that a mapping with a MusicBee field is not present for that enum

    ALL = 0
    NAME = 1000

    # tags/core properties
    TITLE = 65
    ARTIST = 32
    ARTISTS = 108
    ALBUM = 30  # MusicBee album ignoring articles like 'the' and 'a' etc.
    ALBUM_ARTIST = 31
    TRACK_NUMBER = 86
    TRACK_TOTAL = 87
    GENRES = 59
    DATE = 900  # no MusicBee mapping
    YEAR = 35
    MONTH = 901  # no MusicBee mapping
    DAY = 902  # no MusicBee mapping
    BPM = 85
    KEY = 903  # unknown MusicBee mapping
    DISC_NUMBER = 52
    DISC_TOTAL = 54
    COMPILATION = 904  # unknown MusicBee mapping
    COMMENTS = 44
    IMAGES = 905  # no MusicBee mapping
    LENGTH = 16
    RATING = 75
    COMPOSER = 43
    CONDUCTOR = 45
    PUBLISHER = 73

    # file properties
    PATH = 106
    FOLDER = 179
    FILENAME = 3
    EXT = 100
    SIZE = 7
    TYPE = 4
    CHANNELS = 8
    BIT_RATE = 10
    BIT_DEPTH = 183
    SAMPLE_RATE = 9

    # date properties
    DATE_CREATED = 921  # no MusicBee mapping
    DATE_MODIFIED = 11
    DATE_ADDED = 12
    LAST_PLAYED = 13

    # miscellaneous properties
    PLAY_COUNT = 14
    DESCRIPTION = 931  # no MusicBee mapping

    # remote properties
    URI = 941  # no MusicBee mapping
    USER_ID = 942  # no MusicBee mapping
    USER_NAME = 943  # no MusicBee mapping
    OWNER_ID = 944  # no MusicBee mapping
    OWNER_NAME = 945  # no MusicBee mapping
    FOLLOWERS = 946  # no MusicBee mapping


###########################################################################
## Tags
###########################################################################
@dataclass(frozen=True)
class TagMap:
    """Map of human-friendly tag name to ID3 tag ids for a file type"""
    # helpful for determining tags map: https://wiki.hydrogenaud.io/index.php?title=Tag_Mapping

    title: Sequence[str] = field(default=())
    artist: Sequence[str] = field(default=())
    album: Sequence[str] = field(default=())
    album_artist: Sequence[str] = field(default=())
    track_number: Sequence[str] = field(default=())
    track_total: Sequence[str] = field(default=())
    genres: Sequence[str] = field(default=())
    date: Sequence[str] = field(default=())
    year: Sequence[str] = field(default=())
    month: Sequence[str] = field(default=())
    day: Sequence[str] = field(default=())
    bpm: Sequence[str] = field(default=())
    key: Sequence[str] = field(default=())
    disc_number: Sequence[str] = field(default=())
    disc_total: Sequence[str] = field(default=())
    compilation: Sequence[str] = field(default=())
    comments: Sequence[str] = field(default=())
    images: Sequence[str] = field(default=())

    def __getitem__(self, key: str) -> Sequence[str]:
        """Safely get the value of a given attribute key, returning an empty string if the key is not found"""
        return getattr(self, key, [])


class TagField(Field):
    """Applies extra functionality to :py:class:`Field` for objects which contain modifiable tags"""

    __tags__: frozenset[str] = frozenset(list(TagMap.__annotations__.keys()) + ["uri"])

    @classmethod
    def all(cls, only_tags: bool = False) -> list[Self]:
        """
        Get all enums for this enum.
        When ``only_tags`` is True, returns only those enums that represent a tag for this TagField type.
        """
        enums = super().all()
        if not only_tags:
            return enums

        return list(sorted(cls.from_name(*cls.to_tags(enums)), key=lambda x: enums.index(x)))

    def to_tag(self) -> set[str]:
        """
        Returns all human-friendly tag names for the current enum value.

        Applies mapper to enums before returning as per :py:meth:`map`.
        This will only return tag names if they are found in :py:class:`TagMap`.
        """
        if self == Fields.ALL:
            return {tag.name.lower() for tag in super().all() if tag.name.lower() in self.__tags__}
        return {tag.name.lower() for tag in self.map(self) if tag.name.lower() in self.__tags__}

    @classmethod
    def to_tags(cls, tags: UnitIterable[Self]) -> set[str]:
        """
        Returns all human-friendly tag names for the given enum value.

        Applies mapper to enums before returning as per :py:meth:`map`.
        This will only return tag names if they are found in :py:class:`TagMap`.
        """
        if isinstance(tags, cls):
            return tags.to_tag()
        return {t for tag in tags for t in tag.to_tag()}


class TagFields(TagField):
    """
    All possible TagField enums in this program.

    This is used to ensure all TagField enum implementations have the same values for their enum names.
    """
    ALL = Fields.ALL.value
    NAME = Fields.NAME.value

    # tags/core properties
    TITLE = Fields.TITLE.value
    ARTIST = Fields.ARTIST.value
    ALBUM = Fields.ALBUM.value
    ALBUM_ARTIST = Fields.ALBUM_ARTIST.value
    TRACK_NUMBER = Fields.TRACK_NUMBER.value
    TRACK_TOTAL = Fields.TRACK_TOTAL.value
    GENRES = Fields.GENRES.value
    DATE = Fields.DATE.value
    YEAR = Fields.YEAR.value
    MONTH = Fields.MONTH.value
    DAY = Fields.DAY.value
    BPM = Fields.BPM.value
    KEY = Fields.KEY.value
    DISC_NUMBER = Fields.DISC_NUMBER.value
    DISC_TOTAL = Fields.DISC_TOTAL.value
    COMPILATION = Fields.COMPILATION.value
    COMMENTS = Fields.COMMENTS.value
    IMAGES = Fields.IMAGES.value
    LENGTH = Fields.LENGTH.value
    RATING = Fields.RATING.value
    COMPOSER = Fields.COMPOSER.value
    CONDUCTOR = Fields.CONDUCTOR.value
    PUBLISHER = Fields.PUBLISHER.value

    # file properties
    PATH = Fields.PATH.value
    FOLDER = Fields.FOLDER.value
    FILENAME = Fields.FILENAME.value
    EXT = Fields.EXT.value
    SIZE = Fields.SIZE.value
    TYPE = Fields.TYPE.value
    CHANNELS = Fields.CHANNELS.value
    BIT_RATE = Fields.BIT_RATE.value
    BIT_DEPTH = Fields.BIT_DEPTH.value
    SAMPLE_RATE = Fields.SAMPLE_RATE.value

    # date properties
    DATE_CREATED = Fields.DATE_CREATED.value
    DATE_MODIFIED = Fields.DATE_MODIFIED.value
    DATE_ADDED = Fields.DATE_ADDED.value
    LAST_PLAYED = Fields.LAST_PLAYED.value

    # miscellaneous properties
    PLAY_COUNT = Fields.PLAY_COUNT.value
    DESCRIPTION = Fields.DESCRIPTION.value

    # remote properties
    URI = Fields.URI.value
    USER_ID = Fields.USER_ID.value
    USER_NAME = Fields.USER_NAME.value
    OWNER_ID = Fields.OWNER_ID.value
    OWNER_NAME = Fields.OWNER_NAME.value
    FOLLOWERS = Fields.FOLLOWERS.value


ALL_FIELDS = frozenset(Fields.all())
ALL_TAG_FIELDS = frozenset(TagFields.all())


###########################################################################
## Object-specific fields
###########################################################################
class TrackFieldMixin(TagField):
    """Applies extra functionality to the TagField enum for TagField types relating to :py:class:`Track` types"""

    # noinspection PyUnresolvedReferences
    @classmethod
    def map(cls, enum: Self) -> list[Self]:
        if enum == cls.TRACK:
            return [cls.TRACK_NUMBER, cls.TRACK_TOTAL]
        elif enum == cls.DISC:
            return [cls.DISC_NUMBER, cls.DISC_TOTAL]
        elif enum == cls.DATE:
            return [cls.DATE, cls.YEAR, cls.MONTH, cls.DAY]
        return [enum]


class TrackField(TrackFieldMixin):
    """Represents all currently supported fields for objects of type :py:class:`Track`"""
    ALL = TagFields.ALL.value

    TITLE = TagFields.TITLE.value
    ARTIST = TagFields.ARTIST.value
    ALBUM = TagFields.ALBUM.value
    ALBUM_ARTIST = TagFields.ALBUM_ARTIST.value
    TRACK = TagFields.TRACK_NUMBER.value + 500
    TRACK_NUMBER = TagFields.TRACK_NUMBER.value
    TRACK_TOTAL = TagFields.TRACK_TOTAL.value
    GENRES = TagFields.GENRES.value
    DATE = TagFields.DATE.value
    YEAR = TagFields.YEAR.value
    MONTH = TagFields.MONTH.value
    DAY = TagFields.DAY.value
    BPM = TagFields.BPM.value
    KEY = TagFields.KEY.value
    DISC = TagFields.DISC_NUMBER.value + 500
    DISC_NUMBER = TagFields.DISC_NUMBER.value
    DISC_TOTAL = TagFields.DISC_TOTAL.value
    COMPILATION = TagFields.COMPILATION.value
    COMMENTS = TagFields.COMMENTS.value
    IMAGES = TagFields.IMAGES.value
    LENGTH = TagFields.LENGTH.value
    RATING = TagFields.RATING.value

    # remote properties
    URI = TagFields.URI.value


class PlaylistField(Field):
    """Represents all currently supported fields for objects of type :py:class:`Playlist`"""
    ALL = Fields.ALL.value

    # tags/core properties
    TRACK_TOTAL = Fields.TRACK_TOTAL.value
    IMAGES = Fields.IMAGES.value
    LENGTH = Fields.LENGTH.value

    # date properties
    DATE_CREATED = Fields.DATE_CREATED.value
    DATE_MODIFIED = Fields.DATE_MODIFIED.value

    # miscellaneous properties
    DESCRIPTION = Fields.DESCRIPTION.value


class FolderField(Field):
    """Represents all currently supported fields for objects of type :py:class:`Folder`"""
    ALL = Fields.ALL.value

    # tags/core properties
    TRACK_TOTAL = Fields.TRACK_TOTAL.value
    GENRES = Fields.GENRES.value
    IMAGES = Fields.IMAGES.value
    COMPILATION = Fields.COMPILATION.value
    LENGTH = Fields.LENGTH.value

    # file properties
    FOLDER = Fields.FOLDER.value


class AlbumField(Field):
    """Represents all currently supported fields for objects of type :py:class:`Album`"""
    ALL = Fields.ALL.value

    # tags/core properties
    ARTIST = Fields.ARTIST.value
    ALBUM = Fields.ALBUM.value
    ALBUM_ARTIST = Fields.ALBUM_ARTIST.value
    TRACK_TOTAL = Fields.TRACK_TOTAL.value
    GENRES = Fields.GENRES.value
    DATE = TagFields.DATE.value
    YEAR = TagFields.YEAR.value
    MONTH = TagFields.MONTH.value
    DAY = TagFields.DAY.value
    DISC_TOTAL = Fields.DISC_TOTAL.value
    COMPILATION = Fields.COMPILATION.value
    IMAGES = Fields.IMAGES.value
    LENGTH = Fields.LENGTH.value
    RATING = Fields.RATING.value


class ArtistField(Field):
    """Represents all currently supported fields for objects of type :py:class:`Artist`"""
    ALL = Fields.ALL.value

    # tags/core properties
    ARTIST = Fields.ARTIST.value
    TRACK_TOTAL = Fields.TRACK_TOTAL.value
    GENRES = Fields.GENRES.value
    IMAGES = Fields.IMAGES.value
    LENGTH = Fields.LENGTH.value
    RATING = Fields.RATING.value



"""
Operations relating to mapping and re-mapping of paths.
"""
import os
from collections.abc import Collection, Iterable
from os import sep
from pathlib import Path
from typing import Any

from musify.file.base import File
from musify.printer import PrettyPrinter

type PathInputType = str | Path | File | None


class PathMapper(PrettyPrinter):
    """
    Simple path mapper which extracts paths from :py:class:`File` objects.
    Can be extended by child classes for more complex mapping operations.
    """

    __slots__ = ()

    def map(self, value: PathInputType, check_existence: bool = False) -> str | None:
        """
        Map the given ``value`` by either extracting the path from a :py:class:`File` object,
        or returning the ``value`` as is, assuming it is a string.

        :param value: The value to extract a path from.
        :param check_existence: When True, check the path exists before returning it. If it doesn't exist, returns None.
        :return: The path if ``check_existence`` is False, or if ``check_existence`` is True and path exists,
            None otherwise.
        """
        if not value:
            return

        path = str(value.path if isinstance(value, File) else value)
        if not check_existence or os.path.exists(path):
            return path

    def map_many(self, values: Collection[PathInputType], check_existence: bool = False) -> list[str]:
        """Run :py:meth:`map` operation on many ``values`` only returning those values that are not None or empty."""
        paths = [self.map(value=value, check_existence=check_existence) for value in values]
        return [path for path in paths if path]

    def unmap(self, value: PathInputType, check_existence: bool = False) -> str | None:
        """
        Map the given ``value`` by either extracting the path from a :py:class:`File` object,
        or returning the ``value`` as is, assuming it is a string.

        :param value: The value to extract a path from.
        :param check_existence: When True, check the path exists before returning it. If it doesn't exist, returns None.
        :return: The path if ``check_existence`` is False, or if ``check_existence`` is True and path exists,
            None otherwise.
        """
        if not value:
            return

        path = str(value.path if isinstance(value, File) else value)
        if not check_existence or os.path.exists(path):
            return path

    def unmap_many(self, values: Collection[PathInputType], check_existence: bool = False) -> list[str]:
        """Run :py:meth:`unmap` operation on many ``values`` only returning those values that are not None or empty."""
        paths = [self.unmap(value=value, check_existence=check_existence) for value in values]
        return [path for path in paths if path]

    def as_dict(self) -> dict[str, Any]:
        return {}


class PathStemMapper(PathMapper):
    """
    A more complex path mapper which attempts to replace the stems of paths from strings and :py:class:`File` objects.
    Plus, attempts to case-correct paths.

    Useful for cross-platform support. Can be used to correct paths if the same file exists in
    different locations according to different mounts and/or multiple operating systems.
    """

    __slots__ = ("_stem_map", "_stem_unmap", "_available_paths")

    @property
    def available_paths(self) -> dict[str, str]:
        """
        A map of the available paths stored in this object. Simply ``{<lower-case path>: <correctly-cased path>}``.
        When assigning new values to this property, the stored map will update itself
        with the new values rather than overwrite.
        """
        return self._available_paths

    @available_paths.setter
    def available_paths(self, values: Iterable[str | Path]):
        self._available_paths.update({path.casefold(): path for path in map(str, values)})

    @property
    def stem_map(self) -> dict[str, str]:
        """
        A map of ``{<stem to be replaced>: <its replacement>}``.
        Assigning new values to this property updates itself
        plus the ``stem_unmap`` property with the reverse of this map.
        """
        return self._stem_map

    @stem_map.setter
    def stem_map(self, values: dict[str | Path, str | Path]):
        self._stem_map.update({str(k): str(v) for k, v in values.items()})

    @property
    def stem_unmap(self) -> dict[str, str]:
        """
        A map of ``{<replacement stems>: <stem to be replaced>}`` i.e. just the opposite map of ``stem_map``.
        Assign new values to ``stem_map`` to update.
        """
        return {v: k for k, v in self._stem_map.items()}

    def __init__(
            self,
            stem_map: dict[str | Path, str | Path] | None = None,
            available_paths: Iterable[str | Path] = ()
    ):
        self._stem_map: dict[str, str] = {}
        self._available_paths: dict[str, str] = {}

        self.stem_map = stem_map or {}
        self.available_paths = available_paths

    def map(self, value: PathInputType, check_existence: bool = False) -> str | None:
        """
        Map the given value by replacing its stem according to stored ``stem_map``,
        correcting path separators according to the separators of the replacement stem,
        and case correcting path from stored ``available_paths``.

        :param value: The value to map.
        :param check_existence: When True, check the path exists before returning it. If it doesn't exist, returns None.
        :return: The path if ``check_existence`` is False, or if ``check_existence`` is True and path exists,
            None otherwise.
        """
        if not value:
            return

        path = str(value.path if isinstance(value, File) else value)

        seps = ()
        for stem, replacement in self.stem_map.items():
            if path.casefold().startswith(stem.casefold()):
                if "/" in replacement and "/" not in path:
                    seps = ("\\", "/")
                elif "\\" in replacement and "\\" not in path:
                    seps = ("/", "\\")
                path = sep.join([replacement.rstrip("\\/"), path[len(stem):].lstrip("\\/")]).rstrip("\\/")
                break

        if sep == "\\":
            path = path.replace("\\\\", "\\")
        if seps:
            path = path.replace(*seps)

        path = self.available_paths.get(path.casefold(), path)
        if not check_existence or os.path.exists(path):
            return path

    def unmap(self, value: PathInputType, check_existence: bool = False) -> str | None:
        """
        Map the given value by replacing its stem according to stored ``stem_unmap``,
        correcting path separators according to the separators of the replacement stem,
        and case correcting path from stored ``available_paths`` (i.e. mostly the reverse of :py:meth:`map`).

        :param value: The value to map.
        :param check_existence: When True, check the path exists before returning it. If it doesn't exist, returns None.
        :return: The path if ``check_existence`` is False, or if ``check_existence`` is True and path exists,
            None otherwise.
        """
        if not value:
            return

        path = str(value.path if isinstance(value, File) else value)

        seps = ()
        for stem, replacement in self.stem_unmap.items():
            if path.casefold().startswith(stem.casefold()):
                if "/" in replacement and "/" not in path:
                    seps = ("\\", "/")
                elif "\\" in replacement and "\\" not in path:
                    seps = ("/", "\\")
                path = sep.join([replacement.rstrip("\\/"), path[len(stem):].lstrip("\\/")])
                break

        if sep == "\\":
            path = path.replace("\\\\", "\\")
        if seps:
            path = path.replace(*seps)

        path = self.available_paths.get(path.casefold(), path)
        if not check_existence or os.path.exists(path):
            return path

    def as_dict(self) -> dict[str, Any]:
        return {
            "stem_map": self.stem_map,
            "stem_unmap": self.stem_unmap,
            "available_paths": list(self.available_paths.values())
        }



"""
Implements endpoints for manipulating playlists with the Spotify API.
"""
from abc import ABCMeta
from collections.abc import Sequence, Mapping
from itertools import batched
from typing import Any

from yarl import URL

from musify import PROGRAM_NAME, PROGRAM_URL
from musify.libraries.remote.core import RemoteResponse
from musify.libraries.remote.core.exception import APIError, RemoteIDTypeError
from musify.libraries.remote.core.types import APIInputValueSingle, RemoteIDType, RemoteObjectType
from musify.libraries.remote.spotify.api.base import SpotifyAPIBase
from musify.utils import limit_value


class SpotifyAPIPlaylists(SpotifyAPIBase, metaclass=ABCMeta):
    """API endpoints for processing collections i.e. playlists, albums, shows, and audiobooks"""

    __slots__ = ()

    async def load_user_playlists(self) -> None:
        """Load and store user playlists data for the currently authorised user in this API object"""
        responses = await self.get_user_items(kind=RemoteObjectType.PLAYLIST)
        self.user_playlist_data = {response["name"]: response for response in responses}

    async def get_playlist_url(self, playlist: APIInputValueSingle[RemoteResponse]) -> URL:
        """
        Determine the type of the given ``playlist`` and return its API URL.
        If type cannot be determined, attempt to find the playlist in the
        list of the currently authorised user's loaded playlists.

        If you find this method is giving unexpected results when giving the name of a playlist.
        You may reload the currently loaded user's playlists by calling :py:meth:`load_user_playlists`

        :param playlist: One of the following to identify the playlist URL:
            - playlist URL/URI/ID,
            - the name of the playlist in the current user's playlists,
            - the API response of a playlist.
            - a RemoteResponse object representing a remote playlist.
        :return: The playlist URL.
        :raise RemoteIDTypeError: Raised when the function cannot determine the item type of
            the input ``playlist``. Or when it does not recognise the type of the input ``playlist`` parameter.
        """
        if isinstance(playlist, RemoteResponse):
            playlist = playlist.response

        if isinstance(playlist, Mapping):
            if self.url_key in playlist:
                url = playlist[self.url_key]
            elif self.id_key in playlist:
                url = self.wrangler.convert(
                    playlist[self.id_key],
                    kind=RemoteObjectType.PLAYLIST,
                    type_in=RemoteIDType.ID,
                    type_out=RemoteIDType.URL
                )
            elif "uri" in playlist:
                url = self.wrangler.convert(
                    playlist["uri"],
                    kind=RemoteObjectType.PLAYLIST,
                    type_in=RemoteIDType.URI,
                    type_out=RemoteIDType.URL
                )
            else:
                raise APIError(f"Could not determine URL from given input: {playlist}")

            return URL(url)

        try:
            url = self.wrangler.convert(playlist, kind=RemoteObjectType.PLAYLIST, type_out=RemoteIDType.URL)
        except RemoteIDTypeError:
            if not self.user_playlist_data:
                await self.load_user_playlists()

            if playlist not in self.user_playlist_data:
                raise RemoteIDTypeError(
                    "Given playlist is not a valid URL/URI/ID and name not found in user's playlists",
                    value=playlist
                )

            url = self.user_playlist_data[playlist][self.url_key]

        return URL(url)

    ###########################################################################
    ## POST endpoints
    ###########################################################################
    async def create_playlist(
            self, name: str, public: bool = True, collaborative: bool = False, *_, **__
    ) -> dict[str, Any]:
        """
        ``POST: /users/{user_id}/playlists`` - Create an empty playlist for the current user with the given name.

        :param name: Name of playlist to create.
        :param public: Set playlist availability as `public` if True and `private` if False.
        :param collaborative: Set playlist to collaborative i.e. other users may edit the playlist.
        :return: API JSON response for the created playlist.
        """
        url = f"{self.url}/users/{self.user_id}/playlists"

        body = {
            "name": name,
            "description": f"Generated using {PROGRAM_NAME}: {PROGRAM_URL}",
            "public": public,
            "collaborative": collaborative,
        }
        response = (await self.handler.post(url, json=body))
        name = response["name"]
        url = response[self.url_key]
        self.user_playlist_data[name] = response

        # response from creating playlist gives back incorrect user info on 'owner' key, fix it
        response["owner"]["display_name"] = self.user_name
        response["owner"][self.id_key] = self.user_id
        response["owner"]["uri"] = self.wrangler.convert(
            self.user_id, kind=RemoteObjectType.USER, type_in=RemoteIDType.ID, type_out=RemoteIDType.URI
        )
        response["owner"][self.url_key] = self.wrangler.convert(
            self.user_id, kind=RemoteObjectType.USER, type_in=RemoteIDType.ID, type_out=RemoteIDType.URL
        )
        response["owner"]["external_urls"][self.source.lower()] = self.wrangler.convert(
            self.user_id, kind=RemoteObjectType.USER, type_in=RemoteIDType.ID, type_out=RemoteIDType.URL_EXT
        )

        self.handler.log("DONE", url, message=f"Created playlist: {name!r} -> {url}")
        return response

    async def add_to_playlist(
            self,
            playlist: APIInputValueSingle[RemoteResponse],
            items: Sequence[str],
            limit: int = 100,
            skip_dupes: bool = True
    ) -> int:
        """
        ``POST: /playlists/{playlist_id}/tracks`` - Add list of tracks to a given playlist.

        :param playlist: One of the following to identify the playlist to add to:
            - playlist URL/URI/ID,
            - the name of the playlist in the current user's playlists,
            - the API response of a playlist.
            - a RemoteResponse object representing a remote playlist.
        :param items: List of URLs/URIs/IDs of the tracks to add.
        :param limit: Size of each batch of IDs to add. This value will be limited to be between ``1`` and ``100``.
        :param skip_dupes: Skip duplicates.
        :return: The number of tracks added to the playlist.
        :raise RemoteIDTypeError: Raised when the input ``playlist`` does not represent
            a playlist URL/URI/ID.
        :raise RemoteObjectTypeError: Raised when the item types of the input ``items``
            are not all tracks or IDs.
        """
        url = f"{await self.get_playlist_url(playlist)}/tracks"

        if len(items) == 0:
            self.handler.log("SKIP", url, message="No data given")
            return 0

        self.wrangler.validate_item_type(items, kind=RemoteObjectType.TRACK)

        uri_list = [
            self.wrangler.convert(item, kind=RemoteObjectType.TRACK, type_out=RemoteIDType.URI) for item in items
        ]
        if skip_dupes:  # skip tracks currently in playlist
            pl_current = next(iter(await self.get_items(url, kind=RemoteObjectType.PLAYLIST)))
            tracks_key = self.collection_item_map[RemoteObjectType.PLAYLIST].name.lower() + "s"
            tracks = pl_current[tracks_key][self.items_key]

            uri_current = [track["track"]["uri"] for track in tracks]
            uri_list = [uri for uri in uri_list if uri not in uri_current]

        limit = limit_value(limit, floor=1, ceil=100)
        for uris in batched(uri_list, limit):  # add tracks in batches
            await self.handler.post(url, json={"uris": uris}, log_message=f"Adding {len(uris):>6} items")

        self.handler.log("DONE", url, message=f"Added {len(uri_list):>6} items to playlist: {url}")
        return len(uri_list)

    ###########################################################################
    ## PUT endpoints
    ###########################################################################
    async def follow_playlist(self, playlist: APIInputValueSingle[RemoteResponse], *args, **kwargs) -> URL:
        url = URL(f"{await self.get_playlist_url(playlist)}/followers")
        await self.handler.put(url)
        return url

    ###########################################################################
    ## DELETE endpoints
    ###########################################################################
    async def delete_playlist(self, playlist: APIInputValueSingle[RemoteResponse]) -> URL:
        """
        ``DELETE: /playlists/{playlist_id}/followers`` - Unfollow a given playlist.
        WARNING: This function will destructively modify your remote playlists.

        :param playlist: One of the following to identify the playlist to delete:
            - playlist URL/URI/ID,
            - the name of the playlist in the current user's playlists,
            - the API response of a playlist.
            - a RemoteResponse object representing a remote playlist.
        :return: API URL for playlist.
        """
        url = URL(f"{await self.get_playlist_url(playlist)}/followers")
        await self.handler.delete(url)
        return url

    async def clear_from_playlist(
            self,
            playlist: APIInputValueSingle[RemoteResponse],
            items: Sequence[str] | None = None,
            limit: int = 100
    ) -> int:
        """
        ``DELETE: /playlists/{playlist_id}/tracks`` - Clear tracks from a given playlist.
        WARNING: This function can destructively modify your remote playlists.

        :param playlist: One of the following to identify the playlist to clear from :
            - playlist URL/URI/ID,
            - the name of the playlist in the current user's playlists,
            - the API response of a playlist.
            - a RemoteResponse object representing a remote playlist.
        :param items: List of URLs/URIs/IDs of the tracks to remove. If None, clear all songs from the playlist.
        :param limit: Size of each batch of IDs to clear in a single request.
            This value will be limited to be between ``1`` and ``100``.
        :return: The number of tracks cleared from the playlist.
        :raise RemoteIDTypeError: Raised when the input ``playlist`` does not represent
            a playlist URL/URI/ID.
        :raise RemoteObjectTypeError: Raised when the item types of the input ``items``
            are not all tracks or IDs.
        """
        url = f"{await self.get_playlist_url(playlist)}/tracks"
        if items is not None and len(items) == 0:
            self.handler.log("SKIP", url, message="No data given")
            return 0

        if items is None:  # clear everything
            pl_current = next(iter(await self.get_items(url, kind=RemoteObjectType.PLAYLIST, extend=True)))

            tracks_key = self.collection_item_map[RemoteObjectType.PLAYLIST].name.lower() + "s"
            tracks = pl_current[tracks_key][self.items_key]
            uri_list = [track[tracks_key.rstrip("s")]["uri"] for track in tracks]
        else:  # clear only the items given
            self.wrangler.validate_item_type(items, kind=RemoteObjectType.TRACK)
            uri_list = [
                self.wrangler.convert(item, kind=RemoteObjectType.TRACK, type_out=RemoteIDType.URI) for item in items
            ]

        if not uri_list:  # skip when nothing to clear
            self.handler.log("SKIP", url, message="No tracks to clear")
            return 0

        async def _delete_batch(batch: list[str]) -> None:
            body = {"tracks": [{"uri": uri} for uri in batch]}
            await self.handler.delete(url,  json=body, log_message=f"Clearing {len(uri_list):>3} tracks")

        limit = limit_value(limit, floor=1, ceil=100)
        await self.logger.get_asynchronous_iterator(map(_delete_batch, batched(uri_list, limit)), disable=True)

        self.handler.log("DONE", url, message=f"Cleared {len(uri_list):>3} tracks")
        return len(uri_list)



"""
Utilities for operations on :py:class:`LocalPlaylist` types.

Generally, this will contain global variables representing all supported playlist file types
and a utility function for loading the appropriate :py:class:`LocalPlaylist` type for a path based on its extension.
"""
from collections.abc import Collection
from pathlib import Path

from musify.file.exception import InvalidFileType
from musify.file.path_mapper import PathMapper
from musify.libraries.local.playlist.base import LocalPlaylist
from musify.libraries.local.playlist.m3u import M3U
from musify.libraries.local.playlist.xautopf import XAutoPF, REQUIRED_MODULES as REQUIRED_XAUTOPF_MODULES
from musify.libraries.local.track import LocalTrack
from musify.libraries.remote.core.wrangle import RemoteDataWrangler
from musify.utils import required_modules_installed

_playlist_classes = {M3U}
if required_modules_installed(REQUIRED_XAUTOPF_MODULES):
    _playlist_classes.add(XAutoPF)

PLAYLIST_CLASSES = frozenset(_playlist_classes)
PLAYLIST_FILETYPES = frozenset(filetype for c in PLAYLIST_CLASSES for filetype in c.valid_extensions)


async def load_playlist(
        path: str | Path,
        tracks: Collection[LocalTrack] = (),
        path_mapper: PathMapper = PathMapper(),
        remote_wrangler: RemoteDataWrangler = None,
) -> LocalPlaylist:
    """
    Attempt to load a file from a given path, returning the appropriate :py:class:`LocalPlaylist` object

    :param path: Absolute path of the playlist.
        If the playlist ``path`` given does not exist, the playlist instance will use all the tracks
        given in ``tracks`` as the tracks in the playlist.
    :param tracks: Optional. Available Tracks to search through for matches.
        If no tracks are given, the playlist instance will load all the tracks from scratch according to its settings.
    :param path_mapper: Optionally, provide a :py:class:`PathMapper` for paths stored in the playlist file.
        Useful if the playlist file contains relative paths and/or paths for other systems that need to be
        mapped to absolute, system-specific paths to be loaded and back again when saved.
    :param remote_wrangler: Optionally, provide a :py:class:`RemoteDataWrangler` object for processing URIs on tracks.
        If given, the wrangler can be used when calling __get_item__ to get an item from the collection from its URI.
        The wrangler is also used when loading tracks to allow them to process URI tags.
        For more info on this, see :py:class:`LocalTrack`.
    :return: Loaded :py:class:`LocalPlaylist` object
    :raise InvalidFileType: If the file type is not supported.
    """
    ext = Path(path).suffix
    if ext not in PLAYLIST_FILETYPES:
        raise InvalidFileType(ext, f"Not an accepted extension. Use only: {', '.join(PLAYLIST_FILETYPES)}")

    cls = next(cls for cls in PLAYLIST_CLASSES if ext in cls.valid_extensions)
    playlist = cls(path=path, path_mapper=path_mapper, remote_wrangler=remote_wrangler)
    return await playlist.load(tracks=tracks)



"""
Functionality relating to generic remote objects.

Implements core :py:class:`MusifyItem` and :py:class:`MusifyCollection` classes for remote object types.
"""
from __future__ import annotations

from abc import ABCMeta, abstractmethod
from collections.abc import Iterable
from dataclasses import dataclass
from datetime import datetime
from typing import Self, Literal

from musify.base import MusifyItem, Result
from musify.libraries.core.collection import MusifyCollection
from musify.libraries.core.object import Track, Album, Playlist, Artist
from musify.libraries.remote.core.api import RemoteAPI
from musify.libraries.remote.core.base import RemoteObject, RemoteItem
from musify.libraries.remote.core.exception import RemoteError, APIError
from musify.libraries.remote.core.types import APIInputValueSingle
from musify.utils import get_most_common_values


class RemoteTrack(Track, RemoteItem, metaclass=ABCMeta):
    """Extracts key ``track`` data from a remote API JSON response."""

    __slots__ = ()
    __attributes_classes__ = (Track, RemoteItem)


class RemoteCollection[T: RemoteObject](MusifyCollection[T], metaclass=ABCMeta):
    """Generic class for storing a collection of remote objects."""

    __slots__ = ()
    __attributes_classes__ = MusifyCollection
    __attributes_ignore__ = ("items", "track_total")


class RemoteCollectionLoader[T: RemoteObject](RemoteCollection[T], RemoteItem, metaclass=ABCMeta):
    """Generic class for storing a collection of remote objects that can be loaded from an API response."""

    __slots__ = ()
    __attributes_classes__ = (RemoteObject, RemoteCollection)

    def __eq__(self, __collection: RemoteObject | MusifyCollection | Iterable[T]):
        if isinstance(__collection, RemoteObject):
            return self.uri == __collection.uri
        return super().__eq__(__collection)

    @property
    @abstractmethod
    def _total(self) -> int:
        """The total expected items for this collection"""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    async def load(
            cls, value: APIInputValueSingle[Self], api: RemoteAPI, items: Iterable[T] = (), *args, **kwargs
    ) -> Self:
        """
        Generate a new object, calling all required endpoints to get a complete set of data for this item type.

        ``value`` may be:
            * A string representing a URL/URI/ID.
            * A remote API JSON response for a collection with a valid ID value under an ``id`` key.
            * An object of the same type as this collection.
              The remote API JSON response will be used to load a new object.

        You may also provide a set of kwargs relating that will extend aspects of the response
        before using it to initialise a new object. See :py:meth:`reload` for possible extensions.

        :param value: The value representing some remote collection. See description for allowed value types.
        :param api: An authorised API object to load the object from.
        :param items: Optionally, give a list of available items to build a response for this collection.
            In doing so, the method will first try to find the API responses for the items of this collection
            in the given list before calling the API for any items not found there.
            This helps reduce the number of API calls made on initialisation.
        """
        raise NotImplementedError

    def _check_total(self) -> None:
        """
        Checks the total tracks processed for this collection equal to the collection total, raise exception if not
        """
        if self._total != len(self.items):
            raise RemoteError(
                f"{self.name} | "
                "The total items available in the response does not equal the total item count for this collection. "
                "Make sure the given collection response contains the right number of item responses: "
                f"{self._total} != {len(self.items)}"
            )


@dataclass(frozen=True)
class SyncResultRemotePlaylist(Result):
    """Stores the results of a sync with a remote playlist."""
    #: The total number of tracks in the playlist before the sync.
    start: int
    #: The number of tracks added to the playlist.
    added: int
    #: The number of tracks removed from the playlist.
    removed: int
    #: The number of tracks that were in the playlist before and after the sync.
    unchanged: int
    #: The difference between the total number tracks in the playlist from before and after the sync.
    difference: int
    #: The total number of tracks in the playlist after the sync.
    final: int


PLAYLIST_SYNC_KINDS = Literal["new", "refresh", "sync"]


class RemotePlaylist[T: RemoteTrack](Playlist[T], RemoteCollectionLoader[T], metaclass=ABCMeta):
    """Extracts key ``playlist`` data from a remote API JSON response."""

    __slots__ = ()
    __attributes_classes__ = (Playlist, RemoteCollectionLoader)

    @property
    @abstractmethod
    def owner_id(self) -> str:
        """The ID of the owner of this playlist"""
        raise NotImplementedError

    @property
    @abstractmethod
    def owner_name(self) -> str:
        """The name of the owner of this playlist"""
        raise NotImplementedError

    @property
    @abstractmethod
    def followers(self) -> int:
        """The number of followers this playlist has"""
        raise NotImplementedError

    @property
    @abstractmethod
    def date_added(self) -> dict[str, datetime]:
        """A map of ``{<URI>: <date>}`` for each item for when that item was added to the playlist"""
        raise NotImplementedError

    @property
    def _total(self):
        return self.track_total

    @property
    def writeable(self) -> bool:
        """Is this playlist writeable i.e. can this program modify it"""
        try:
            self._check_for_api()
        except APIError:
            return False
        return self.api.user_id == self.owner_id

    @classmethod
    async def create(cls, api: RemoteAPI, name: str, public: bool = True, collaborative: bool = False) -> Self:
        """
        Create an empty playlist for the current user with the given name
        and initialise and return a new RemotePlaylist object from this new playlist.

        :param api: An API object with authorised access to a remote User to create playlists for.
        :param name: Name of playlist to create.
        :param public: Set playlist availability as `public` if True and `private` if False.
        :param collaborative: Set playlist to collaborative i.e. other users may edit the playlist.
        :return: :py:class:`RemotePlaylist` object for the generated playlist.
        """
        response = await api.create_playlist(name=name, public=public, collaborative=collaborative)
        return cls(response=response, api=api)

    async def delete(self) -> None:
        """
        Unfollow/delete the current playlist and clear the stored response for this object.
        WARNING: This function will destructively modify your remote playlists.
        """
        self._check_for_api()
        await self.api.delete_playlist(self.url)
        self.response.clear()

    async def sync(
            self,
            items: Iterable[MusifyItem] = (),
            kind: PLAYLIST_SYNC_KINDS = "new",
            reload: bool = True,
            dry_run: bool = True,
    ) -> SyncResultRemotePlaylist:
        """
        Synchronise this playlist object's items with the remote playlist it is associated with.

        Sync options:
            * 'new': Do not clear any items from the remote playlist and only add any tracks
                from this playlist object not currently in the remote playlist.
            * 'refresh': Clear all items from the remote playlist first, then add all items from this playlist object.
            * 'sync': Clear all items not currently in this object's items list, then add all tracks
                from this playlist object not currently in the remote playlist.

        :param items: Provide an item collection or list of items to synchronise to the remote playlist.
            Use the currently loaded ``tracks`` in this object if not given.
        :param kind: Sync option for the remote playlist. See description.
        :param reload: When True, once synchronisation is complete, reload this RemotePlaylist object
            to reflect the changes on the remote playlist if enabled. Skip if False.
        :param dry_run: Run function, but do not modify the remote playlists at all.
        :return: The results of the sync as a :py:class:`SyncResultRemotePlaylist` object.
        """
        if not self.writeable:
            raise RemoteError(f"Cannot write to this playlist: {self.name}")

        uri_initial = [track.uri for track in (items or self.tracks) if track.uri]
        uri_remote = self._get_track_uris_from_api_response()

        # default settings when only synchronising for new items
        uri_add = [uri for uri in uri_initial if uri not in uri_remote]
        uri_unchanged = uri_remote
        removed = 0

        # process the remote playlist. when dry_run, mock the results
        if kind == "refresh":  # remove all items from the remote playlist
            removed = await self.api.clear_from_playlist(self.url) if not dry_run else len(uri_remote)
            uri_add = uri_initial
            uri_unchanged = []
        elif kind == "sync":  # remove items not present in the current list from the remote playlist
            uri_clear = [uri for uri in uri_remote if uri not in uri_initial]
            removed = await self.api.clear_from_playlist(self.url, items=uri_clear) if not dry_run else len(uri_clear)
            uri_unchanged = [uri for uri in uri_remote if uri in uri_initial]

        added = len(uri_add)
        if not dry_run:
            added = await self.api.add_to_playlist(self.url, items=uri_add, skip_dupes=kind != "refresh")
            if reload:  # reload the current playlist object from remote
                await self.reload(extend_tracks=True)

        return SyncResultRemotePlaylist(
            start=len(uri_remote),
            added=added,
            removed=removed,
            unchanged=len(uri_unchanged),
            difference=len(self.tracks) - len(uri_remote) if not dry_run and reload else added - removed,
            final=len(self.tracks) if not dry_run and reload else len(uri_remote) + added - removed
        )

    @abstractmethod
    def _get_track_uris_from_api_response(self) -> list[str]:
        """
        Returns a list of URIs from the API response stored for this playlist
        Implementation of this method is needed for the :py:func:`sync` function.
        """
        raise NotImplementedError


class RemoteAlbum[T: RemoteTrack](Album[T], RemoteCollectionLoader[T], metaclass=ABCMeta):
    """Extracts key ``album`` data from a remote API JSON response."""

    __slots__ = ()
    __attributes_classes__ = (Album, RemoteCollectionLoader)

    @property
    def _total(self):
        return self.track_total

    @property
    @abstractmethod
    def artists(self) -> list[RemoteArtist[T]]:
        raise NotImplementedError


class RemoteArtist[T: (RemoteTrack, RemoteAlbum)](Artist[T], RemoteCollectionLoader[T], metaclass=ABCMeta):
    """Extracts key ``artist`` data from a remote API JSON response."""

    __slots__ = ()
    __attributes_classes__ = (Artist, RemoteCollectionLoader)

    @property
    def _total(self):
        return 0

    @property
    def tracks(self) -> list[T]:
        return [track for album in self.albums for track in album]

    @property
    def artists(self):
        return get_most_common_values(artist.name for album in self.albums for artist in album.artists)

    @property
    @abstractmethod
    def albums(self) -> list[RemoteAlbum[T]]:
        raise NotImplementedError

    @property
    def track_total(self):
        return sum(album.track_total for album in self.albums)

    @property
    @abstractmethod
    def image_links(self) -> dict[str, str]:
        """The images associated with this artist in the form ``{<image name/type>: <image link>}``"""
        raise NotImplementedError

    @property
    def has_image(self) -> bool:
        """Does this artist have images associated with them"""
        return len(self.image_links) > 0

    @property
    def length(self):
        if self.albums and any(album.length for album in self.albums):
            return sum(album.length if album.length else 0 for album in self.albums)



"""
Exceptions relating to file operations.
"""
from pathlib import Path

from musify.exception import MusifyError


class FileError(MusifyError):
    """
    Exception raised for file errors.

    :param file: The file type that caused the error.
    :param message: Explanation of the error.
    """

    def __init__(self, file: str | Path | None = None, message: str | None = None):
        self.file = Path(file)
        self.message = message
        formatted = f"{file} | {message}" if file else message
        super().__init__(formatted)


class InvalidFileType(FileError):
    """
    Exception raised for unrecognised file types.

    :param filetype: The file type that caused the error.
    :param message: Explanation of the error.
    """

    def __init__(self, filetype: str, message: str = "File type not recognised"):
        self.message = message
        super().__init__(file=filetype, message=message)


class FileDoesNotExistError(FileError, FileNotFoundError):
    """
    Exception raised when a file cannot be found.

    :param path: The path that caused the error.
    :param message: Explanation of the error.
    """

    def __init__(self, path: str | Path, message: str = "File cannot be found"):
        self.message = message
        super().__init__(file=path, message=message)


class UnexpectedPathError(FileError):
    """
    Exception raised when a path is invalid.
    Usually raised when a directory is given when a file was expected and vice versa.

    :param path: The path that caused the error.
    :param message: Explanation of the error.
    """

    def __init__(self, path: str | Path, message: str = "Invalid path given"):
        self.message = message
        super().__init__(file=path, message=message)


class ImageLoadError(FileError):
    """Exception raised for errors in loading an image."""



"""
Just the core abstract class for the :py:mod:`Remote` module.
Placed here separately to avoid circular import logic issues.
"""
from abc import ABCMeta, abstractmethod
from typing import Any

from yarl import URL

from musify.base import MusifyObject


class RemoteResponse(MusifyObject, metaclass=ABCMeta):

    __slots__ = ()

    # noinspection PyPropertyDefinition,PyMethodParameters
    @property
    @abstractmethod
    def kind(cls):
        """The type of remote object this class represents"""
        raise NotImplementedError

    @property
    @abstractmethod
    def response(self) -> dict[str, Any]:
        """The API response for this object"""
        raise NotImplementedError

    @property
    @abstractmethod
    def id(self) -> str:
        """The ID of this item/collection."""
        raise NotImplementedError

    @property
    @abstractmethod
    def url(self) -> URL:
        """The API URL of this item/collection."""
        raise NotImplementedError

    @property
    @abstractmethod
    def url_ext(self) -> URL | None:
        """The external URL of this item/collection."""
        raise NotImplementedError

    @abstractmethod
    def refresh(self, skip_checks: bool = False) -> None:
        """
        Refresh this object by updating from the stored API response.
        Useful for updating stored variables after making changes to the stored API response manually.
        """
        raise NotImplementedError



"""
Base functionality to be shared by all classes that implement :py:class:`RemoteAPI` functionality for Spotify.
"""
from abc import ABCMeta
from collections.abc import Collection, MutableMapping, Iterable
from typing import Any

from aiorequestful.auth.oauth2 import AuthorisationCodeFlow
from aiorequestful.cache.backend.base import ResponseRepository
from aiorequestful.cache.exception import CacheError
from aiorequestful.cache.session import CachedSession
from aiorequestful.types import URLInput
from yarl import URL

from musify.libraries.remote.core.api import RemoteAPI
from musify.libraries.remote.core.types import RemoteObjectType


class SpotifyAPIBase(RemoteAPI[AuthorisationCodeFlow], metaclass=ABCMeta):
    """Base functionality required for all endpoint functions for the Spotify API"""

    __slots__ = ()

    #: The key to reference when extracting items from a collection
    items_key = "items"

    ###########################################################################
    ## Format values/responses
    ###########################################################################
    @staticmethod
    def _format_key(key: str | RemoteObjectType | None) -> str | None:
        """Get the expected key in a response from a :py:class:`RemoteObjectType`"""
        if key is None:
            return
        if isinstance(key, RemoteObjectType):
            key = key.name
        return key.lower().rstrip("s") + "s"

    @staticmethod
    def format_next_url(url: URLInput, offset: int = 0, limit: int = 20) -> str:
        """Format a `next` style URL for looping through API pages"""
        url = URL(url)

        params: dict[str, Any] = dict(url.query)
        params["offset"] = offset
        params["limit"] = limit

        url = url.with_query(params)
        return str(url)

    ###########################################################################
    ## Enrich/manipulate responses
    ###########################################################################
    def _enrich_with_identifiers(self, response: dict[str, Any], id_: str, href: str) -> None:
        """Ensure key identifiers are present in the response."""
        if self.id_key not in response:
            response[self.id_key] = id_
        if self.url_key not in response:
            response[self.url_key] = href

    def _enrich_with_parent_response(
            self,
            response: MutableMapping[str, Any],
            key: str,
            parent_key: str | RemoteObjectType | None,
            parent_response: MutableMapping[str, Any]
    ) -> None:
        """
        Some endpoints don't include parent response on child items.
        This method adds the given ``parent_response`` back to the child ``response`` on the given ``key``.
        """
        if (
                not parent_key
                or isinstance(parent_key, str)
                or parent_key == RemoteObjectType.PLAYLIST
                or self.items_key not in response
                or self.items_key in parent_response
        ):
            return

        parent_key_name = self._format_key(parent_key).rstrip("s")
        parent_response = {k: v for k, v in parent_response.items() if k != key}

        if not parent_response:
            return

        for item in response[self.items_key]:
            if parent_key_name not in item:
                item[parent_key_name] = parent_response

    ###########################################################################
    ## Cache utilities
    ###########################################################################
    async def _get_responses_from_cache(
            self, method: str, url: URLInput, id_list: Collection[str]
    ) -> tuple[list[dict[str, Any]], Collection[str], Collection[str]]:
        """
        Attempt to find the given ``id_list`` in the cache of the request handler and return results.

        :param url: The base API URL endpoint for the required requests.
        :param id_list: List of IDs to append to the given URL.
        :return: (Results from the cache, IDs found in the cache, IDs not found in the cache)
        """
        session = self.handler.session
        if not isinstance(session, CachedSession):
            self.handler.log("CACHE", url, message="Cache not configured, skipping...")
            return [], [], id_list

        repository = session.cache.get_repository_from_url(url=url)
        if repository is None:
            self.handler.log("CACHE", url, message="No repository for this endpoint, skipping...")
            return [], [], id_list

        results = await repository.get_responses([(method.upper(), id_,) for id_ in id_list])
        ids_found = {result[self.id_key] for result in results}
        ids_not_found = {id_ for id_ in id_list if id_ not in ids_found}

        self.handler.log(
            method="CACHE",
            url=url,
            message=[f"Retrieved {len(results):>6} cached responses", f"{len(ids_not_found):>6} not found in cache"]
        )
        return results, ids_found, ids_not_found

    async def _cache_responses(self, method: str, responses: Iterable[dict[str, Any]]) -> None:
        """Persist ``results`` of a given ``method`` to the cache."""
        session = self.handler.session
        if not isinstance(session, CachedSession) or not responses:
            return

        # take all parts of href path, excluding ID
        possible_urls = {"/".join(result.get(self.url_key, "").split("/")[:-1]) for result in responses}
        possible_urls = {url for url in possible_urls if url}
        if not possible_urls:
            return
        if len(possible_urls) > 1:
            raise CacheError(
                "Too many different types of results given. Given results must relate to the same repository type."
            )
        results_mapped = {(method.upper(), result[self.id_key]): result for result in responses}
        url = next(iter(possible_urls))
        repository: ResponseRepository = session.cache.get_repository_from_url(url)
        if repository is not None:
            self.handler.log(
                method="CACHE",
                url=url,
                message=f"Caching {len(results_mapped)} responses to {repository.settings.name!r} repository",
            )
            await repository.save_responses({k: await repository.serialize(v) for k, v in results_mapped.items()})



"""
Processor making comparisons between objects and data types.
"""
import re
from collections.abc import Sequence
from datetime import datetime, date
from functools import reduce
from operator import mul
from typing import Any

from musify.base import MusifyItem
from musify.field import Field
from musify.processors.base import DynamicProcessor, dynamicprocessormethod
from musify.processors.exception import ComparerError
from musify.processors.time import TimeMapper
from musify.types import UnitSequence
from musify.utils import to_collection


class Comparer(DynamicProcessor):
    """
    Compares an item or object with another item, object or a given set of expected values to find a match.

    :param condition: The condition to match on e.g. Is, LessThan, InRange, Contains.
    :param expected: Optional list of expected values to match on.
        Types of the values in this list are automatically converted to the type of the item field's value.
    :param field: The field to match on.
    :param reference_required: When True, a reference object of type ``T`` must be passed to the ``compare`` method.
        When False, reference files given to the ``compare`` method will be ignored.
        An exception will be raised if this is True and reference object is not passed.
    """

    __slots__ = ("_expected", "_converted", "field", "reference_required")

    @classmethod
    def _processor_method_fmt(cls, name: str) -> str:
        return "_" + cls._pascal_to_snake(name)

    @property
    def condition(self) -> str:
        """String representation of the current condition name of this object"""
        return self._processor_name.lstrip("_")

    @property
    def expected(self) -> list[Any] | None:
        """A list of expected values used for most conditions"""
        return self._expected

    @expected.setter
    def expected(self, value: Sequence[Any] | None):
        """Set the list of expected values and reset the ``_converted`` attribute to False"""
        self._converted = False
        self._expected = to_collection(value, list)

    def __init__(
            self,
            condition: str,
            expected: UnitSequence[Any] | None = None,
            field: Field | None = None,
            reference_required: bool = False,
    ):
        super().__init__()
        self._expected: list[Any] | None = None
        self._converted = False

        self.expected: list[Any] | None = to_collection(expected, list)
        #: The :py:class:`Field` representing the property to extract the comparison value from
        #: when an :py:class:`MusifyItem` is given
        self.field: Field | None = field.map(field)[0] if field else None
        #: Whether to raise an exception when :py:meth:`compare` is called and a reference object is not provided.
        self.reference_required = reference_required

        self._set_processor_name(condition)

    def __call__(self, *args, **kwargs) -> bool:
        return self.compare(*args, **kwargs)

    def compare[T: Any](self, item: T, reference: T | None = None) -> bool:
        """
        Compare a ``item`` to a ``reference`` or,
        if no ``reference`` is given, to this object's list of ``expected`` values

        :return: True if a match is found, False otherwise.
        :raise LocalProcessorError: If no reference given and no expected values set for this comparer.
        """
        if self.condition is None:
            return False

        if reference is None and self.reference_required:
            raise ComparerError("A reference is required for this instance of Comparer")
        if reference is None and not self.expected:
            raise ComparerError("No comparative item given and no expected values set")

        tag_name = None
        if self.field and isinstance(item, MusifyItem):
            tag_name = self.field.name.lower()
            actual = item[tag_name]
        else:
            actual = item

        if self.reference_required:  # use the values from the reference as the expected values
            expected = to_collection(reference[tag_name], list)
        else:  # convert the expected values to the same type as the actual value if not yet converted
            if not self._converted:
                self._convert_expected(actual)
            expected = self.expected

        if expected:  # special on-the-fly conversions for datetime values
            if isinstance(actual, datetime) and not isinstance(expected[0], datetime):
                actual = actual.date()
            elif not isinstance(actual, datetime) and isinstance(expected[0], datetime):
                expected = [exp.date() for exp in expected]

        return super().__call__(actual, expected)

    def _convert_expected(self, value: Any) -> None:
        """Driver for converting expected values to the same type as given value"""
        if self._converted:
            return

        if isinstance(value, int):
            self._convert_expected_to_int()
            self._converted = True
        elif isinstance(value, float):
            self._convert_expected_to_float()
            self._converted = True
        elif isinstance(value, datetime):
            self._convert_expected_to_datetime()
            self._converted = True
        elif isinstance(value, bool):
            self._expected.clear()
            self._converted = True
        elif isinstance(value, str):
            self._converted = True

    def _convert_expected_to_int(self) -> None:
        """Convert expected values to integers"""
        converted: list[int] = []
        for exp in self.expected:
            if isinstance(exp, str) and ":" in exp:
                # value is a string representation of time
                exp = self._get_seconds(exp)
            converted.append(int(exp))
        self._expected = converted

    def _convert_expected_to_float(self) -> None:
        """Convert expected values to floats"""
        converted: list[float] = []
        for exp in self.expected:
            if isinstance(exp, str) and ":" in exp:
                # value is a string representation of time
                exp = self._get_seconds(exp)
            converted.append(float(exp))
        self._expected = converted

    def _convert_expected_to_datetime(self) -> None:
        """Convert expected values to :py:class:`datetime` objects"""
        converted: list[date] = []

        for exp in self.expected:
            if isinstance(exp, datetime):
                converted.append(exp.date())
            elif re.match(r"\d{1,2}[/-]\d{1,2}[/-]\d{2,4}", exp):
                # value is a string representation of datetime
                digits: list[int] = list(map(int, exp.split("/")[::-1]))

                if len(str(digits[-1])) < 4:  # year is not fully qualified, add millennium part
                    this_millennium = str(date.today().year)[:2]
                    last_millennium = str(int(this_millennium) - 1)

                    if digits[0] % 1000 > date.today().year % 1000:
                        digits[0] = int(last_millennium + str(digits[0])[-2:].zfill(2))
                    else:
                        digits[0] = int(this_millennium + str(digits[0])[-2:].zfill(2))

                converted.append(date(*digits))
            else:  # value is durational difference, calculate datetime using the current time
                digit = int(re.sub(r"\D+", "", exp))
                mapper_key = re.sub(r"\d+", "", exp)
                converted.append(datetime.now() - TimeMapper(mapper_key)(digit))

        self._expected = converted

    @staticmethod
    def _get_seconds(time_str: str) -> float:
        """Convert string representation of time to seconds e.g. 4:30 -> 270s"""
        factors = (24, 60, 60, 1)
        digits_split = time_str.split(":")
        digits = tuple(int(n.split(",")[0]) for n in digits_split)

        seconds = 0
        if "," in digits_split[-1]:  # add milliseconds if present
            seconds += int(digits_split[-1].split(",")[1]) / 1000

        for i, digit in enumerate(reversed(digits), 1):  # convert to seconds
            seconds += digit * reduce(mul, factors[-i:], 1)

        return seconds

    @dynamicprocessormethod
    def _is(self, value: Any | None, expected: Sequence[Any] | None) -> bool:
        return value == expected[0]

    @dynamicprocessormethod
    def _is_not(self, value: Any | None, expected: Sequence[Any] | None) -> bool:
        return not self._is(value=value, expected=expected)

    @dynamicprocessormethod("greater_than", "in_the_last")
    def _is_after(self, value: Any | None, expected: Sequence[Any] | None) -> bool:
        return value > expected[0] if value is not None and expected[0] is not None else False

    @dynamicprocessormethod("less_than", "not_in_the_last")
    def _is_before(self, value: Any | None, expected: Sequence[Any] | None) -> bool:
        return value < expected[0] if value is not None and expected[0] is not None else False

    @dynamicprocessormethod
    def _is_in(self, value: Any | None, expected: Sequence[Any] | None) -> bool:
        return value in expected

    @dynamicprocessormethod
    def _is_not_in(self, value: Any | None, expected: Sequence[Any] | None) -> bool:
        return not self._is_in(value=value, expected=expected)

    @dynamicprocessormethod
    def _in_range(self, value: Any | None, expected: Sequence[Any] | None) -> bool:
        return expected[0] <= value <= expected[1] if value is not None and expected[0] is not None else False

    @dynamicprocessormethod
    def _not_in_range(self, value: Any | None, expected: Sequence[Any] | None) -> bool:
        return not self._in_range(value=value, expected=expected)

    @dynamicprocessormethod
    def _is_not_null(self, value: Any | None, _: Sequence[Any] | None = None) -> bool:
        return value is not None or value is True

    @dynamicprocessormethod
    def _is_null(self, value: Any | None, _: Sequence[Any] | None = None) -> bool:
        return value is None or value is False

    @dynamicprocessormethod
    def _starts_with(self, value: Any | None, expected: Sequence[Any] | None) -> bool:
        return value.startswith(str(expected[0])) if value is not None and expected[0] is not None else False

    @dynamicprocessormethod
    def _ends_with(self, value: Any | None, expected: Sequence[Any] | None) -> bool:
        return value.endswith(str(expected[0])) if value is not None and expected[0] is not None else False

    @dynamicprocessormethod
    def _contains(self, value: Any | None, expected: Sequence[Any] | None) -> bool:
        return str(expected[0]) in value if value is not None and expected[0] is not None else False

    @dynamicprocessormethod
    def _does_not_contain(self, value: Any | None, expected: Sequence[Any] | None) -> bool:
        return not self._contains(value=value, expected=expected)

    @dynamicprocessormethod
    def _matches_reg_ex(self, value: Any | None, expected: Sequence[Any] | None) -> bool:
        return bool(re.search(str(expected[0]), value)) if value is not None and expected[0] is not None else False

    @dynamicprocessormethod
    def _matches_reg_ex_ignore_case(self, value: Any | None, expected: Sequence[Any] | None) -> bool:
        if value is not None and expected[0] is not None:
            return False
        return bool(re.search(str(expected[0]), str(value), flags=re.I))

    def as_dict(self):
        return {
            "condition": self.condition,
            "expected": self.expected,
            "field": self.field.name if self.field else None,
            "reference_required": self.reference_required,
        }



"""
Processor that matches objects and data types based on given configuration.
"""
import inspect
import logging
import re
from collections.abc import Iterable, Callable, MutableSequence
from concurrent.futures import ThreadPoolExecutor, Future, Executor
from dataclasses import dataclass, field
from typing import Any

from musify.base import MusifyObject
from musify.field import TagField, TagFields as Tag, ALL_TAG_FIELDS
from musify.libraries.core.collection import MusifyCollection
from musify.logger import MusifyLogger
from musify.printer import PrettyPrinter
from musify.processors.base import Processor
from musify.types import UnitIterable
from musify.utils import limit_value, to_collection


@dataclass
class CleanTagConfig(PrettyPrinter):
    """Config for processing string-type tag values before matching with :py:class:`ItemMatcher`"""
    #: The name of the tag to clean.
    tag: TagField
    #: A set of string values to remove from this tag.
    remove: set[str] = field(default_factory=lambda: set())
    #: A set of string values for which the cleaner will slice the tag on and remove anything that comes after.
    split: set[str] = field(default_factory=lambda: set())
    #: A function to apply before the remove and split values are applied.
    preprocess: Callable[[str], str] = field(default=lambda x: x)

    def as_dict(self) -> dict[str, Any]:
        return {
            "tag": self.tag.name.lower(),
            "remove": [value for value in self.remove],
            "split": [value for value in self.split],
            "preprocess": self.preprocess is not None,
        }


class ItemMatcher(Processor):
    """Matches source items/collections to given result(s)."""

    __slots__ = ("logger",)

    #: A set of words to search for in tag values that identify the item as being a karaoke item.
    karaoke_tags = {"karaoke", "backing", "instrumental"}
    #: A difference in years of this value gives a score of 0 for the :py:meth:`match_year` algorithm.
    #: See the :py:meth:`match_year` method for more information.
    year_range = 10

    # config for cleaning string-type tags for matching
    #: Apply these remove settings to all tags when processing tags as per :py:meth:`clean_tags` method.
    #: See also :py:class:`CleanTagConfig` for more info on this configuration.
    clean_tags_remove_all = {"the", "a", "&", "and"}
    #: Apply these split settings to all tags when processing tags as per :py:meth:`clean_tags` method.
    #: See also :py:class:`CleanTagConfig` for more info on this configuration.
    clean_tags_split_all = set()
    #: A list of configurations in the form of :py:class:`CleanTagConfig`
    #: to apply for each tag type. See also :py:class:`CleanTagConfig` for more info.
    clean_tags_config = (
        CleanTagConfig(tag=Tag.TITLE, remove={"part"}, split={"featuring", "feat.", "ft.", "/"}),
        CleanTagConfig(tag=Tag.ARTIST, split={"featuring", "feat.", "ft.", "vs"}),
        CleanTagConfig(tag=Tag.ALBUM, remove={"ep"}, preprocess=lambda x: x.split('-')[0])
    )

    # config for name score reduction
    #: A set of words to check for when applying name score reduction logic.
    #:
    #: If a word from this list is present in the name of the result to score against
    #: but not in the source :py:class:`MusifyObject`, apply the ``reduce_name_score_factor`` to reduce its score.
    #: This set is always combined with the ``karaoke_tags``.
    reduce_name_score_on = {"live", "demo", "acoustic"}
    #: The factor to apply to a name score when a word from ``reduce_name_score_on``
    #: is found in the result but not in the source :py:class:`MusifyObject`.
    reduce_name_score_factor = 0.5

    def __init__(self):
        # noinspection PyTypeChecker
        #: The :py:class:`MusifyLogger` for this  object
        self.logger: MusifyLogger = logging.getLogger(__name__)

    def log(self, messages: MutableSequence[str], pad: str = ' ') -> None:
        """
        Log lists of ``messages`` in a uniform aligned format with a given ``pad`` character.

        Convenience function for ensuring consistent log format for results of operations of this class
        and any other classes which use this class.
        """
        messages[0] = pad * 3 + ' ' + (messages[0] if messages[0] else "unknown")
        self.logger.debug(" | ".join(messages))

    def _log_algorithm(self, source: MusifyObject, extra: Iterable[str] = ()) -> None:
        """Wrapper for initially logging an algorithm in a uniform aligned format"""
        algorithm = inspect.stack()[1][0].f_code.co_name.upper().lstrip("_").replace("_", " ")
        log = [source.name, algorithm]
        if extra:
            log.extend(extra)
        self.log(log, pad='>')

    def _log_test[T: MusifyObject](self, source: T, result: T | None, test: Any, extra: Iterable[str] = ()) -> None:
        """Wrapper for initially logging a test result in a uniform aligned format"""
        algorithm = inspect.stack()[1][0].f_code.co_name.replace("match", "").upper().lstrip("_").replace("_", " ")

        if result is not None and hasattr(result, "uri"):
            log_result = f"> Testing URI: {result.uri}"
        else:
            log_result = "> Test failed"

        log = [source.name, log_result, f"{algorithm:<10}={test:<5}"]
        if extra:
            log.extend(extra)
        self.log(log)

    def _log_match[T: MusifyObject](self, source: T, result: T, extra: Iterable[str] = ()) -> None:
        """Wrapper for initially logging a match in a correctly aligned format"""
        log = [source.name, f"< Matched URI: {result.uri}"]
        if extra:
            log.extend(extra)
        self.log(log, pad='<')

    def clean_tags(self, source: MusifyObject) -> None:
        """
        Clean tags on the input item and assign to its ``clean_tags`` attribute. Used for better matching/searching.
        Clean by removing words, and only taking phrases before a certain word e.g. 'featuring', 'part'.
        Cleaning config for string-type tags is set in ``_clean_tags_config``.

        :param source: The base object with tags to clean.
        """
        def process(val: str, conf: CleanTagConfig) -> str:
            """Apply transformations to the given ``val`` to clean it"""
            val = conf.preprocess(val)
            val = re.sub(r"[(\[].*?[)\]]", "", val).casefold()

            for word in conf.remove | self.clean_tags_remove_all:
                val = re.sub(rf"\s{word}\s|^{word}\s|\s{word}$", " ", val)

            for word in conf.split | self.clean_tags_split_all:
                val = val.split(word)[0].rstrip()

            return re.sub(r"[^\w']+", ' ', val).strip()

        source.clean_tags.clear()
        source.clean_tags[Tag.NAME] = ""
        name = source.name

        # process string tags according to config
        for config in self.clean_tags_config:
            tag_name = config.tag.to_tag().pop()
            value = getattr(source, tag_name, None)
            if not value:
                source.clean_tags[config.tag] = ""
                continue

            value_cleaned = process(value, conf=config)
            source.clean_tags[config.tag] = value_cleaned
            if name == value:
                source.clean_tags[Tag.NAME] = value_cleaned

        source.clean_tags[Tag.LENGTH] = getattr(source, Tag.LENGTH.name.lower(), None)
        source.clean_tags[Tag.YEAR] = getattr(source, Tag.YEAR.name.lower(), None)

    ###########################################################################
    ## Conditions
    ###########################################################################
    def match_not_karaoke[T: MusifyObject](self, source: T, result: T) -> int:
        """Checks if a result is not a karaoke item that is either 0 when item is karaoke or 1 when not karaoke."""
        def is_karaoke(*values: str) -> bool:
            """Check if the words in the given ``values`` match any word in ``karaoke_tags``"""
            values = {v for value in values for v in value.casefold().split()}
            karaoke = any(word.casefold() in values for word in self.karaoke_tags)
            self._log_test(source=source, result=result, test=karaoke, extra=[f"{self.karaoke_tags} -> {values}"])
            return karaoke

        if is_karaoke(result.name):
            return 0
        if is_karaoke(result.artist):
            return 0
        if is_karaoke(result.album):
            return 0
        return 1

    def match_name[T: MusifyObject](self, source: T, result: T) -> float:
        """Match on names and return a score between 0-1. Score=0 when either value is None."""
        score = 0
        source_val = source.clean_tags.get(Tag.NAME)
        result_val = result.clean_tags.get(Tag.NAME)

        if source_val and result_val:
            score = sum(word in result_val for word in source_val.split()) / len(source_val.split())

            # reduce a score if certain keywords are present in result and not source
            reduce_on = self.reduce_name_score_on | self.karaoke_tags
            reduce_on = {word.casefold() for word in reduce_on}
            if any(word in result.name.casefold() and word not in source.name.casefold() for word in reduce_on):
                score = max(score * self.reduce_name_score_factor, 0)

        self._log_test(source=source, result=result, test=round(score, 2), extra=[f"{source_val} -> {result_val}"])
        return score

    def match_artist[T: MusifyObject](self, source: T, result: T) -> float:
        """
        Match on artists and return a score between 0-1. Score=0 when either value is None.
        When many artists are present, a scale factor is applied to the score of matches on subsequent artists.
        i.e. match on artist 1 is scaled by 1, match on artist 2 is scaled by 1/2,
        match on artist 3 is scaled by 1/3 etc.
        """
        score = 0
        source_val = source.clean_tags.get(Tag.ARTIST)
        result_val = result.clean_tags.get(Tag.ARTIST)

        if not source_val or not result_val:
            self._log_test(source=source, result=result, test=score, extra=[f"{source_val} -> {result_val}"])
            return score

        artists_source = source_val.replace(MusifyObject.tag_sep, " ")
        artists_result = result_val.split(MusifyObject.tag_sep)

        for i, artist in enumerate(artists_result, 1):
            score += (sum(word in artists_source for word in artist.split()) / len(artists_source.split())) * (1 / i)
        self._log_test(source=source, result=result, test=round(score, 2), extra=[f"{source_val} -> {result_val}"])
        return score

    def match_album[T: MusifyObject](self, source: T, result: T) -> float:
        """Match on album and return a score between 0-1. Score=0 when either value is None."""
        score = 0

        source_val = source.clean_tags.get(Tag.ALBUM)
        result_val = result.clean_tags.get(Tag.ALBUM)

        if source_val and result_val:
            score = sum(word in result_val for word in source_val.split()) / len(source_val.split())

        self._log_test(source=source, result=result, test=round(score, 2), extra=[f"{source_val} -> {result_val}"])
        return score

    def match_length[T: MusifyObject](self, source: T, result: T) -> float:
        """Match on length and return a score between 0-1. Score=0 when either value is None."""
        score = 0
        source_val = source.clean_tags.get(Tag.LENGTH)
        result_val = result.clean_tags.get(Tag.LENGTH)

        if source_val and result_val:
            score = max((source_val - abs(source_val - result_val)), 0) / source_val

        extra = [f"{round(source_val, 2) if source_val else None} -> {round(result_val, 2) if result_val else None}"]
        self._log_test(source=source, result=result, test=round(score, 2), extra=extra)
        return score

    def match_year[T: MusifyObject](self, source: T, result: T) -> float:
        """
        Match on year and return a score between 0-1. Score=0 when either value is None.
        Matches within ``year_range`` years on a 0-1 scale where 1 is the exact same year
        and 0 is a difference in year greater that ``year_range``.
        User may modify this max range via the ``year_range`` class attribute.
        """
        score = 0
        source_val = source.clean_tags.get(Tag.YEAR)
        result_val = result.clean_tags.get(Tag.YEAR)

        if source_val and result_val:
            score = max((self.year_range - abs(source_val - result_val)), 0) / self.year_range

        self._log_test(source=source, result=result, test=round(score, 2), extra=[f"{source_val} -> {result_val}"])
        return score

    ###########################################################################
    ## Score match
    ###########################################################################
    def __call__[T: MusifyObject](self, *args, **kwargs) -> T | None:
        return self.match(*args, **kwargs)

    def match[T: MusifyObject](
            self,
            source: T,
            results: Iterable[T],
            min_score: float = 0.1,
            max_score: float = 0.8,
            match_on: UnitIterable[TagField] = ALL_TAG_FIELDS,
            allow_karaoke: bool = False,
    ) -> T | None:
        """
        Perform match algorithm for a given item and its results.

        :param source: Source item to compare against and find a match for.
        :param results: Results for comparisons.
        :param min_score: Only return the result as a match if the score is above this value.
            Value will be limited to between 0.01 and 1.0.
        :param max_score: Stop matching once this score has been reached.
            Value will be limited to between 0.01 and 1.0.
        :param match_on: List of tags to match on. Currently only the following fields are supported:
            ``title``, ``artist``, ``album``, ``year``, ``length``.
        :param allow_karaoke: When True, items determined to be karaoke are allowed when matching added items.
            Skip karaoke results otherwise. Karaoke items are identified using the ``karaoke_tags`` attribute.
        :return: T. The item that matched best if found, None if no item matched conditions.
        """
        if not source.clean_tags:
            self.clean_tags(source)

        # process and limit match options
        match_on_filtered = set()
        for match_field in to_collection(match_on, set):
            if match_field == Tag.ALL:
                match_on_filtered.update(match_field.all())
            else:
                match_on_filtered.add(match_field)

        min_score = limit_value(min_score, floor=0.01, ceil=1.0)
        max_score = limit_value(max_score, floor=0.01, ceil=1.0)
        self._log_algorithm(source=source, extra=[f"max_score={max_score}"])

        with ThreadPoolExecutor(thread_name_prefix="matcher") as executor:
            scores = self._score(
                source=source,
                results=results,
                executor=executor,
                match_on=match_on_filtered,
                allow_karaoke=allow_karaoke
            )

        result, score = self._get_match_from_scores(scores, max_score=max_score)

        if result is not None and score > min_score:
            extra = [
                f"best score: {score:.2f} > {min_score:.2f}"
                if score < max_score else
                f"max score reached: {score:.2f} > {max_score:.2f}"
            ]
            self._log_match(source=source, result=result, extra=extra)
            return result

        self._log_test(
            source=source, result=result, test=score, extra=[f"NO MATCH: {score:.2f}<{min_score:.2f}"]
        )

    def _score[T: MusifyObject](
            self,
            source: T,
            results: Iterable[T],
            executor: Executor,
            match_on: set[TagField] = ALL_TAG_FIELDS,
            allow_karaoke: bool = False,
    ) -> list[tuple[T, dict[TagField, Future[float] | list[list[Future[float]]]]]]:
        """
        Gets the scores for all given ``results`` against a cleaned ``source``.

        :param source: Source item to compare against and find a match for with assigned ``clean_tags``.
        :param results: Result items for comparisons.
        :param executor: The executor to submit tasks to.
        :param match_on: List of tags to match on. Currently only the following fields are supported:
            ``title``, ``artist``, ``album``, ``year``, ``length``.
        :param allow_karaoke: When True, items determined to be karaoke are allowed when matching added items.
            Skip karaoke results otherwise. Karaoke items are identified using the ``karaoke_tags`` attribute.
        :return: Tuple of (the score between 0-1, the item that had the best score)
        """
        scores: list[tuple[T, dict[TagField, Future[float] | list[Future[float]]]]] = []
        if not results:
            self._log_algorithm(source=source, extra=["NO RESULTS GIVEN, SKIPPING"])
            return scores

        for result in results:
            self.clean_tags(result)
            result_scores = self._get_scores(
                source=source, result=result, executor=executor, match_on=match_on, allow_karaoke=allow_karaoke
            )
            if not result_scores:
                continue
            scores.append((result, result_scores))

        return scores

    def _get_scores[T: MusifyObject](
            self,
            source: T,
            result: T,
            executor: Executor,
            match_on: set[TagField] = ALL_TAG_FIELDS,
            allow_karaoke: bool = False,
    ) -> dict[TagField, Future[float] | list[list[Future[float]]]]:
        """
        Gets the scores from a cleaned source and result to match on.
        When an MusifyCollection is given to match on,
        scores are also calculated for each of the items in the collection.
        Scores are always between 0-1.

        :param source: Source item to compare against and find a match for with assigned ``clean_tags``.
        :param result: Result item to compare against with assigned ``clean_tags``.
        :param executor: The executor to submit tasks to.
        :param match_on: List of tags to match on. Currently only the following fields are supported:
            ``title``, ``artist``, ``album``, ``year``, ``length``.
        :param allow_karaoke: When True, items determined to be karaoke are allowed when matching added items.
            Skip karaoke results otherwise. Karaoke items are identified using the ``karaoke_tags`` attribute.
        :return: Map of score type name to score.
        """
        scores: dict[TagField, Future[float] | list[list[Future[float]]]] = {}
        if not allow_karaoke and self.match_not_karaoke(source, result) < 1:
            return scores

        if Tag.TITLE in match_on:
            scores[Tag.TITLE] = executor.submit(self.match_name, source=source, result=result)
        if Tag.ARTIST in match_on:
            scores[Tag.ARTIST] = executor.submit(self.match_artist, source=source, result=result)
        if Tag.ALBUM in match_on:
            scores[Tag.ALBUM] = executor.submit(self.match_album, source=source, result=result)
        if Tag.LENGTH in match_on:
            scores[Tag.LENGTH] = executor.submit(self.match_length, source=source, result=result)
        if Tag.YEAR in match_on:
            scores[Tag.YEAR] = executor.submit(self.match_year, source=source, result=result)

        if isinstance(source, MusifyCollection) and isinstance(result, MusifyCollection):
            # also score all the items individually in the collection
            scores[Tag.ALL] = []
            for item in source.items:
                self.clean_tags(item)
                item_scores = self._score(
                    source=item, results=result.items, match_on=match_on, allow_karaoke=allow_karaoke, executor=executor
                )
                for _, item_score in item_scores:
                    scores[Tag.ALL].append([score for score in item_score.values() if not isinstance(score, list)])

        return scores

    @staticmethod
    def _get_match_from_scores[T: MusifyObject](
            scores: Iterable[tuple[T, dict[TagField, Future[float] | list[list[Future[float]]]]]],
            max_score: float,
    ) -> tuple[T | None, float]:
        best_result = None
        best_score = 0

        def sum_nested_scores(futures: list[list[Future[float]]]) -> float:
            """Sum the scores from a given list of nested Futures"""
            scores_summed = [sum(score.result() for score in nested) / len(nested) for nested in futures]
            return sum(scores_summed) / len(scores_summed)

        for result, result_scores in scores:
            result_scores = [
                sum_nested_scores(score) if isinstance(score, list) else score.result()
                for score in result_scores.values()
            ]
            score = sum(result_scores) / len(result_scores)
            if score > best_score:
                best_score = score
                best_result = result

            if best_score > max_score:
                break

        return best_result, best_score

    def as_dict(self) -> dict[str, Any]:
        return {
            "clean_tags": {
                "remove_all": self.clean_tags_remove_all,
                "split_all": self.clean_tags_split_all,
                "config": self.clean_tags_config,
            },
            "karaoke_tags": self.karaoke_tags,
            "year_range": self.year_range,
        }
